"""
Various ways to gather trajectories in RL environments.
"""

from .list import mean_total_reward, mean_finished_reward
from .logger import EpisodeLogger
from .norm import RewardNormalizer
from .rollers import Roller, BasicRoller, TruncatedRoller, EpisodeRoller
from .rollout import Rollout, empty_rollout

__all__ = dir()



"""
Learning algorithms for RL.
"""

from .advantages import AdvantageEstimator, GAE
from .a2c import A2C
from .ppo import PPO
from .schedules import TFSchedule, LinearTFSchedule

# Don't import mpi by default since that breaks when
# you don't run your program with `mpirun`.

__all__ = dir()



"""
Ways of running gym environments.
"""

from multiprocessing import Pipe, Array, Process

import cloudpickle
import gym.spaces
import numpy as np

from .base import AsyncEnv, BatchedEnv, BatchedAsyncEnv

def batched_gym_env(env_fns, observation_space=None, num_sub_batches=1, sync=False):
    """
    Create a BatchedEnv that controls a set of Gym
    environments.

    Environments are passed via env_fns, which is a list
    of functions that create environments.
    These functions must be picklable with cloudpickle.

    If the observation space is not specified, the first
    function in env_fns will be called an extra time to
    figure out the observation space.

    On top of the normal BatchedEnv interface, the result
    has action_space and observation_space attributes.
    """
    assert len(env_fns) % num_sub_batches == 0
    if not sync and observation_space is None:
        env = env_fns[0]()
        try:
            observation_space = env.observation_space
        finally:
            env.close()
    sub_batches = []
    batch_size = len(env_fns) // num_sub_batches
    for i in range(num_sub_batches):
        batch_fns = env_fns[i*batch_size : (i+1)*batch_size]
        if sync:
            envs = [fn() for fn in batch_fns]
        else:
            envs = [AsyncGymEnv(fn, observation_space) for fn in batch_fns]
        sub_batches.append(envs)
    if sync:
        return BatchedGymEnv(sub_batches)
    return BatchedAsyncEnv(sub_batches)

class AsyncGymEnv(AsyncEnv):
    """
    An AsyncEnv that controls a Gym environment in a
    subprocess.

    On top of the AsyncEnv interface, the resulting object
    has observation_space and action_space attributes.
    """
    def __init__(self, make_env, observation_space):
        self.observation_space = observation_space
        if isinstance(observation_space, gym.spaces.Box):
            num_elems = len(np.array(observation_space.low).flatten())
            zeros = [0] * num_elems
            self._obs_buf = Array('b', zeros)
        else:
            self._obs_buf = None
        self._pipe, other_end = Pipe()
        self._proc = Process(target=self._worker,
                             args=(other_end,
                                   self._obs_buf,
                                   cloudpickle.dumps(make_env)),
                             daemon=True)
        self._proc.start()
        self._running_cmd = None
        other_end.close()
        self._pipe.send(('action_space', None))
        self.action_space = self._get_response()

    def reset_start(self):
        assert self._running_cmd is None
        self._running_cmd = 'reset'
        self._pipe.send(('reset', None))

    def reset_wait(self):
        assert self._running_cmd == 'reset'
        res = self._decode_observation(self._get_response())
        self._running_cmd = None
        return res

    def step_start(self, action):
        assert self._running_cmd is None
        self._running_cmd = 'step'
        self._pipe.send(('step', action))

    def step_wait(self):
        assert self._running_cmd == 'step'
        obs, rew, done, info = self._get_response()
        obs = self._decode_observation(obs)
        self._running_cmd = None
        return obs, rew, done, info

    def close(self):
        if self._running_cmd == 'reset':
            self.reset_wait()
        elif self._running_cmd == 'step':
            self.step_wait()
        assert self._running_cmd is None
        self._running_cmd = 'close'
        self._pipe.send(('close', None))
        self._proc.join()
        self._pipe.close()

    def _get_response(self):
        """
        Read a value from the response queue and make sure
        it is not an exception.
        """
        try:
            resp_obj = self._pipe.recv()
        except EOFError:
            raise RuntimeError('worker has died')
        if isinstance(resp_obj, BaseException):
            raise RuntimeError('exception on worker') from resp_obj
        return resp_obj

    def _decode_observation(self, obs):
        """
        Get an observation out of an object that the
        worker sent.
        """
        if obs is not None:
            return obs
        obs = np.frombuffer(self._obs_buf.get_obj(), dtype='uint8')
        shape = self.observation_space.low.shape
        return obs.reshape(shape).copy()

    @classmethod
    def _worker(cls, pipe, obs_buf, make_env):
        """
        Entry-point for the sub-process.
        """
        try:
            env = cloudpickle.loads(make_env)()
        except BaseException as exc:
            pipe.recv()
            pipe.send(exc)
            return
        try:
            while True:
                cmd, action = pipe.recv()
                if cmd == 'action_space':
                    pipe.send(env.action_space)
                elif cmd == 'reset':
                    pipe.send(cls._sendable_observation(env.reset(), obs_buf))
                elif cmd == 'step':
                    obs, rew, done, info = env.step(action)
                    if done:
                        obs = env.reset()
                    obs = cls._sendable_observation(obs, obs_buf)
                    pipe.send((obs, rew, done, info))
                elif cmd == 'close':
                    return
                else:
                    raise ValueError('unknown command: ' + cmd)
        except BaseException as exc:
            pipe.send(exc)
        finally:
            env.close()

    @staticmethod
    def _sendable_observation(obs, obs_buf):
        """
        Efficiently communicate the observation to the parent
        process, either via shared memory or by returning a
        picklable object.
        """
        if obs_buf is None:
            return obs
        if not isinstance(obs, np.ndarray):
            obs = np.array(obs)
        if obs.dtype != 'uint8':
            return obs
        dst = obs_buf.get_obj()
        dst_np = np.frombuffer(dst, dtype=obs.dtype).reshape(obs.shape)
        np.copyto(dst_np, obs)
        return None

class BatchedGymEnv(BatchedEnv):
    """
    A BatchedEnv that wraps a bunch of existing Gym
    environments and runs them synchronously.
    """
    def __init__(self, envs):
        self.action_space = envs[0][0].action_space
        self.observation_space = envs[0][0].observation_space
        self.envs = envs
        self._step_actions = [None] * len(self.envs)
        self._states = [None] * len(self.envs)

    @property
    def num_sub_batches(self):
        return len(self.envs)

    @property
    def num_envs_per_sub_batch(self):
        return len(self.envs[0])

    def reset_start(self, sub_batch=0):
        assert self._states[sub_batch] is None
        self._states[sub_batch] = 'reset'

    def reset_wait(self, sub_batch=0):
        assert self._states[sub_batch] == 'reset'
        self._states[sub_batch] = None
        return [env.reset() for env in self.envs[sub_batch]]

    def step_start(self, actions, sub_batch=0):
        assert len(actions) == self.num_envs_per_sub_batch
        assert self._states[sub_batch] is None
        self._states[sub_batch] = 'step'
        self._step_actions[sub_batch] = actions

    def step_wait(self, sub_batch=0):
        assert self._step_actions[sub_batch] is not None
        assert self._states[sub_batch] == 'step'
        self._states[sub_batch] = None
        obses, rews, dones, infos = ([], [], [], [])
        for env, action in zip(self.envs[sub_batch], self._step_actions[sub_batch]):
            obs, rew, done, info = env.step(action)
            if done:
                obs = env.reset()
            obses.append(obs)
            rews.append(rew)
            dones.append(done)
            infos.append(info)
        self._step_actions[sub_batch] = None
        return obses, rews, dones, infos

    def close(self):
        for batch in self.envs:
            for env in batch:
                env.close()



"""
Wrappers for padding spaces.
"""

import gym
import numpy as np

class ObservationPadEnv(gym.ObservationWrapper):
    """
    An environment that zero-pads the observation.

    Supports any Box observation space.
    """
    def __init__(self, env, padded_shape, center=True):
        """
        Create a padded environment.

        Args:
          env: the environment to wrap.
          padded_shape: the shape after padding.
          center: if True, attempt to center the original
            observation in the padded one. Otherwise, put
            the original image at the beginning of the
            padded image (e.g. the top-left corner).
        """
        super(ObservationPadEnv, self).__init__(env)
        self._padded_shape = padded_shape
        self._center = center
        old_space = self.observation_space
        self.observation_space = gym.spaces.Box(low=self._observation(old_space.low),
                                                high=self._observation(old_space.high))

    def _observation(self, observation):
        """
        Pad the observation.
        """
        total_pads = tuple(target - cur for target, cur
                           in zip(self._padded_shape, observation.shape))
        assert all(x >= 0 for x in total_pads)
        if self._center:
            pad_width = []
            for pad in total_pads:
                if pad % 2 == 0:
                    pad_width.append((pad//2, pad//2))
                else:
                    pad_width.append((pad//2, pad//2+1))
        else:
            pad_width = [(0, pad) for pad in total_pads]
        return np.pad(observation, pad_width, 'constant')

class MultiBinaryPadEnv(gym.Wrapper):
    """
    An environment that adds no-op actions to a
    multi-binary action space.

    This is useful for unifying several different
    environments with different controls.
    """
    def __init__(self, env, num_actions):
        """
        Create a padded environment.

        Args:
          env: the environment to wrap.
          num_actions: the number of actions in the padded
            action space.
        """
        super(MultiBinaryPadEnv, self).__init__(env)
        assert num_actions >= env.action_space.n
        self._num_actions = num_actions
        self.action_space = gym.spaces.MultiBinary(num_actions)

    def _step(self, action):
        return self.env.step(action[:self.env.action_space.n])



"""
Helper routines for training algorithms.
"""

import tensorflow as tf

def select_from_batch(advs, batch):
    """
    Take a rollout-shaped list of lists and select the
    indices from the mini-batch.
    """
    indices = zip(batch['rollout_idxs'], batch['timestep_idxs'])
    return [advs[x][y] for x, y in indices]

def select_model_out_from_batch(key, rollouts, batch):
    """
    Select a model_outs key corresponding to the indices
    from the mini-batch.
    """
    vals = [[m[key][0] for m in r.model_outs] for r in rollouts]
    return select_from_batch(vals, batch)

def masked_mean(mask, vals):
    """
    Mask the values and compute the mean of the masked
    elements.
    """
    return masked_sum(mask, vals) / tf.reduce_sum(mask)

def masked_sum(mask, vals):
    """
    Mask the values and compute the sum of the masked
    elements.
    """
    masked = tf.where(tf.equal(mask, 0), tf.zeros_like(vals), vals)
    return tf.reduce_sum(masked)



"""
APIs for categorical (discrete) spaces.
"""

import numpy as np
import tensorflow as tf

from .base import Distribution

class CategoricalSoftmax(Distribution):
    """
    A probability distribution that uses softmax to decide
    between a discrete number of options.
    """
    def __init__(self, num_options, low=0):
        self.num_options = num_options
        self.low = low

    @property
    def out_shape(self):
        return (self.num_options,)

    def to_vecs(self, space_elements):
        res = np.zeros((len(space_elements), self.num_options))
        res[np.arange(len(space_elements)), np.array(space_elements)-self.low] = 1
        return res

    @property
    def param_shape(self):
        return (self.num_options,)

    def sample(self, param_batch):
        dist = softmax(np.array(param_batch))
        cumulative_dist = np.cumsum(dist, axis=-1)
        sampled = np.random.rand(len(param_batch), 1)
        large_enoughs = cumulative_dist > sampled
        return self.low + np.argmax(large_enoughs, axis=-1)

    def mode(self, param_batch):
        return self.low + np.argmax(param_batch, axis=-1)

    def log_prob(self, param_batch, sample_vecs):
        loss_func = tf.nn.softmax_cross_entropy_with_logits
        return tf.negative(loss_func(labels=sample_vecs, logits=param_batch))

    def entropy(self, param_batch):
        log_probs = tf.nn.log_softmax(param_batch)
        probs = tf.exp(log_probs)
        return tf.negative(tf.reduce_sum(log_probs * probs, axis=-1))

    def kl_divergence(self, param_batch_1, param_batch_2):
        log_probs_1 = tf.nn.log_softmax(param_batch_1)
        log_probs_2 = tf.nn.log_softmax(param_batch_2)
        probs = tf.exp(log_probs_1)
        return tf.reduce_sum(probs * (log_probs_1 - log_probs_2), axis=-1)

class NaturalSoftmax(CategoricalSoftmax):
    """
    A softmax distribution with natural gradients through
    log_prob into the parameters.

    The forward outputs are like CategoricalSoftmax.
    However, the gradient through log_prob is artificially
    filled in as the natural gradient.
    """
    def __init__(self, num_options, low=0, epsilon=1e-4):
        super(NaturalSoftmax, self).__init__(num_options, low=low)
        self.epsilon = epsilon

    def log_prob(self, param_batch, sample_vecs):
        log_probs = super(NaturalSoftmax, self).log_prob(param_batch, sample_vecs)
        probs = tf.exp(log_probs) + self.epsilon
        neg_grads = -1 / (self.num_options * probs)
        natural_grads = tf.tile(tf.expand_dims(neg_grads, axis=1), (1, self.num_options))
        natural_grads -= self.num_options * natural_grads * sample_vecs
        dots = tf.reduce_sum(param_batch*tf.stop_gradient(natural_grads), axis=-1)
        return tf.stop_gradient(log_probs) + dots - tf.stop_gradient(dots)

def softmax(param_batch):
    """
    Compute a batched softmax on the minor dimension.
    """
    col_shape = (len(param_batch), 1)
    max_vals = np.reshape(param_batch.max(axis=-1), col_shape)
    unnorm = np.exp(param_batch - max_vals)
    return unnorm / np.reshape(np.sum(unnorm, axis=-1), col_shape)



"""
Stateless neural network models.
"""

import math

import numpy as np
import tensorflow as tf
from tensorflow.contrib.layers import fully_connected # pylint: disable=E0611

from .base import TFActorCritic
from .util import mini_batches, product

# pylint: disable=E1129

class FeedforwardAC(TFActorCritic):
    """
    A base class for any feed-forward actor-critic model.

    Subclasses should set several attributes on init:
      obs_ph: placeholder for observation batch
      actor_out: actor output batch
      critic_out: critic output batch. Should be of shape
        (None,).
    """
    def __init__(self, session, action_dist, obs_vectorizer):
        """
        Construct a feed-forward model.
        """
        super(FeedforwardAC, self).__init__(session, action_dist, obs_vectorizer)

        # Set these in your constructor.
        self.obs_ph = None
        self.actor_out = None
        self.critic_out = None

    def scale_outputs(self, scale):
        """
        Scale the network outputs by the given amount.

        This may be called right after initializing the
        model to help deal with different reward scales.
        """
        self.critic_out *= scale
        self.actor_out *= scale

    @property
    def stateful(self):
        return False

    def start_state(self, batch_size):
        return None

    def step(self, observations, states):
        feed_dict = {
            self.obs_ph: self.obs_vectorizer.to_vecs(observations)
        }
        act, val = self.session.run((self.actor_out, self.critic_out), feed_dict)
        return {
            'action_params': act,
            'actions': self.action_dist.sample(act),
            'states': None,
            'values': val
        }

    def batch_outputs(self):
        mask = tf.ones(tf.shape(self.critic_out))
        return self.actor_out, self.critic_out, mask

    def batches(self, rollouts, batch_size=None):
        obses, rollout_idxs, timestep_idxs = _frames_from_rollouts(rollouts)
        for mini_indices in mini_batches([1]*len(obses), batch_size):
            sub_obses = [obses[i] for i in mini_indices]
            yield {
                'rollout_idxs': np.take(rollout_idxs, mini_indices),
                'timestep_idxs': np.take(timestep_idxs, mini_indices),
                'feed_dict': {
                    self.obs_ph: self.obs_vectorizer.to_vecs(sub_obses)
                }
            }

class MLP(FeedforwardAC):
    """
    A multi-layer perceptron actor-critic model.
    """
    # pylint: disable=R0913,R0914
    def __init__(self,
                 session,
                 action_dist,
                 obs_vectorizer,
                 layer_sizes,
                 activation=tf.nn.relu,
                 actor_init=tf.zeros_initializer(),
                 critic_init=tf.zeros_initializer()):
        """
        Create an MLP model.

        Args:
          session: TF session.
          action_dist: an action Distribution.
          obs_vectorizer: an observation SpaceVectorizer.
          layer_sizes: list of hidden layer sizes.
          activation: the activation function.
          actor_init: initializer for the actor head.
          critic_init: initializer for the critic head.
        """
        super(MLP, self).__init__(session, action_dist, obs_vectorizer)

        in_batch_shape = (None,) + obs_vectorizer.out_shape
        self.obs_ph = tf.placeholder(tf.float32, shape=in_batch_shape)

        # Iteratively generate hidden layers.
        layer_in_size = product(obs_vectorizer.out_shape)
        vectorized_shape = (tf.shape(self.obs_ph)[0], layer_in_size)
        layer_in = tf.reshape(self.obs_ph, vectorized_shape)
        for layer_idx, out_size in enumerate(layer_sizes):
            with tf.variable_scope('layer_' + str(layer_idx)):
                layer_in = fully_connected(layer_in, out_size, activation_fn=activation)

        self.base_out = layer_in

        with tf.variable_scope('actor'):
            out_size = product(action_dist.param_shape)
            actor_out = fully_connected(self.base_out, out_size,
                                        activation_fn=None,
                                        weights_initializer=actor_init)
            batch = tf.shape(actor_out)[0]
            self.actor_out = tf.reshape(actor_out, (batch,) + action_dist.param_shape)

        with tf.variable_scope('critic'):
            critic_out = fully_connected(self.base_out, 1,
                                         activation_fn=None,
                                         weights_initializer=critic_init)
            self.critic_out = tf.reshape(critic_out, (tf.shape(critic_out)[0],))

class CNN(FeedforwardAC):
    """
    A convolutional actor-critic model.

    Based on:
    https://github.com/openai/baselines/blob/699919f1cf2527b184f4445a3758a773f333a1ba/baselines/a2c/policies.py#L91
    """
    # pylint: disable=R0913
    def __init__(self,
                 session,
                 action_dist,
                 obs_vectorizer,
                 input_scale=1/0xff,
                 input_dtype=tf.uint8,
                 actor_init=tf.zeros_initializer(),
                 critic_init=tf.zeros_initializer()):
        """
        Create an CNN model.

        Args:
          session: TF session.
          action_dist: an action Distribution.
          obs_vectorizer: an observation SpaceVectorizer.
          input_scale: factor to make inputs well-scaled.
          input_dtype: data-type for input placeholder.
          actor_init: initializer for the actor head.
          critic_init: initializer for the critic head.
        """
        super(CNN, self).__init__(session, action_dist, obs_vectorizer)

        assert len(obs_vectorizer.out_shape) == 3

        in_batch_shape = (None,) + obs_vectorizer.out_shape
        self.obs_ph = tf.placeholder(input_dtype, shape=in_batch_shape)
        obs_batch = tf.cast(self.obs_ph, tf.float32) * input_scale

        with tf.variable_scope('base'):
            self.base_out = self.base(obs_batch)
        with tf.variable_scope('actor'):
            self.actor_out = self.actor(self.base_out, actor_init)
        with tf.variable_scope('critic'):
            self.critic_out = self.critic(self.base_out, critic_init)

    # pylint: disable=R0201
    def base(self, obs_batch):
        """
        Apply the shared part of the model.
        Return a batch that is fed into actor() and
        critic().
        """
        conv_kwargs = {
            'activation': tf.nn.relu,
            'kernel_initializer': tf.orthogonal_initializer(gain=math.sqrt(2))
        }
        with tf.variable_scope('layer_1'):
            cnn_1 = tf.layers.conv2d(obs_batch, 32, 8, 4, **conv_kwargs)
        with tf.variable_scope('layer_2'):
            cnn_2 = tf.layers.conv2d(cnn_1, 64, 4, 2, **conv_kwargs)
        with tf.variable_scope('layer_3'):
            cnn_3 = tf.layers.conv2d(cnn_2, 64, 3, 1, **conv_kwargs)
        flat_size = np.prod(cnn_3.get_shape()[1:])
        flat_in = tf.reshape(cnn_3, (tf.shape(cnn_3)[0], int(flat_size)))
        return tf.layers.dense(flat_in, 512, **conv_kwargs)

    def actor(self, base, initializer):
        """
        Turn the output from base() into action params.
        """
        out_size = product(self.action_dist.param_shape)
        actor_out = fully_connected(base, out_size,
                                    activation_fn=None,
                                    weights_initializer=initializer)
        batch = tf.shape(actor_out)[0]
        return tf.reshape(actor_out, (batch,) + self.action_dist.param_shape)

    def critic(self, base, initializer):
        """
        Turn the output from base() into values.
        """
        critic_out = fully_connected(base, 1,
                                     activation_fn=None,
                                     weights_initializer=initializer)
        return tf.reshape(critic_out, (tf.shape(critic_out)[0],))

def _frames_from_rollouts(rollouts):
    """
    Flatten out the rollouts and produce a list of
    observations, rollout indices, and timestep indices.

    Does not include trailing observations for truncated
    rollouts.

    For example, [[obs1, obs2], [obs3, obs4, obs5]] would
    become ([obs1, obs2, ..., obs5], [0, 0, 1, 1, 1],
    [0, 1, 0, 1, 2])
    """
    all_obs = []
    rollout_indices = []
    timestep_indices = []
    for rollout_idx, rollout in enumerate(rollouts):
        for timestep_idx, obs in enumerate(rollout.step_observations):
            all_obs.append(obs)
            rollout_indices.append(rollout_idx)
            timestep_indices.append(timestep_idx)
    return all_obs, rollout_indices, timestep_indices



"""
Various models for Reinforcement Learning agents.
"""

from .base import Model, TFActorCritic
from .feedforward_ac import FeedforwardAC, MLP, CNN
from .recurrent_ac import RecurrentAC, RNNCellAC

__all__ = dir()



"""
Tests for TensorFlow actor-critic models.
"""

# pylint: disable=E0611
# pylint: disable=E1129

import unittest

import numpy as np
import tensorflow as tf
from tensorflow.contrib.rnn import LSTMCell, MultiRNNCell

from anyrl.models import MLP, RNNCellAC
from anyrl.rollouts import BasicRoller
from anyrl.spaces import gym_space_distribution, gym_space_vectorizer
from anyrl.tests import TupleCartPole

class ModelTester:
    """
    Test a TFActorCritic model.
    """
    def __init__(self, test_case, model_maker, add_noise=True):
        self.test_case = test_case
        self.graph = tf.Graph()
        with self.graph.as_default():
            self.session = tf.Session()
            self.model = model_maker(self.session)
            self.session.run(tf.global_variables_initializer())
            if add_noise:
                for variable in tf.trainable_variables():
                    noise = 0.01 * tf.random_normal(tf.shape(variable))
                    self.session.run(tf.assign(variable, variable + noise))
            self.batch_outputs = self.model.batch_outputs()

    def test_all(self):
        """
        Run all tests.
        """
        with self.session:
            self.test_batches()
            self.test_batched_step()

    def test_batches(self):
        """
        Test that batches() gives consistent and correctly
        structured output.
        """
        with self.graph.as_default():
            for batch_size in [None, 10]:
                for trunc_start in [False, True]:
                    self._test_batches_consistency(batch_size, trunc_start)

    def test_batched_step(self):
        """
        Make sure that the outputs of batched steps are
        the correct shape.
        """
        with self.graph.as_default():
            env = TupleCartPole()
            try:
                obs_space = env.observation_space
            finally:
                env.close()
            batch_size = 7
            in_obses = [obs_space.sample() for _ in range(batch_size)]
            in_states = self.model.start_state(batch_size)
            outs = self.model.step(in_obses, in_states)
            self.test_case.assertEqual(len(outs['actions']), batch_size)
            self.test_case.assertEqual(_state_shape_uid(outs['states']),
                                       _state_shape_uid(in_states))
            if 'action_params' in outs:
                param_shape = self.model.action_dist.param_shape
                self.test_case.assertEqual(np.array(outs['action_params']).shape,
                                           (batch_size,)+param_shape)
            if 'values' in outs:
                self.test_case.assertEqual(np.array(outs['values']).shape,
                                           (batch_size,))

    def _test_batches_consistency(self, batch_size, trunc_start):
        """
        Make sure that batches() produces the same outputs
        that we got with step().
        """
        env = TupleCartPole()
        try:
            roller = BasicRoller(env, self.model, min_episodes=7)
            rollouts = roller.rollouts()
            if trunc_start:
                rollouts = self._truncate_first(rollouts)
            num_batches = 10
            for batch in self.model.batches(rollouts, batch_size=batch_size):
                num_batches -= 1
                if num_batches == 0:
                    break
                self._test_batch(rollouts, batch)
        finally:
            env.close()

    def _test_batch(self, rollouts, batch):
        """
        Test that the batch is consistent with the
        rollouts and with its own mask.
        """
        feed_dict = batch['feed_dict']
        actor_outs, critic_outs, mask = self.session.run(self.batch_outputs,
                                                         feed_dict)
        for i, (action, value) in enumerate(zip(actor_outs, critic_outs)):
            rollout_idx = batch['rollout_idxs'][i]
            timestep_idx = batch['timestep_idxs'][i]
            self.test_case.assertTrue(mask[i] == 0 or mask[i] == 1)
            if mask[i] == 0:
                self.test_case.assertEqual(rollout_idx, 0)
                self.test_case.assertEqual(timestep_idx, 0)
                continue
            self.test_case.assertEqual(mask[i], 1)
            model_out = rollouts[rollout_idx].model_outs[timestep_idx]
            step_action = np.array(model_out['action_params'][0])
            step_value = model_out['values'][0]
            self.test_case.assertEqual(step_action.shape, np.array(action).shape)
            self.test_case.assertEqual(step_action.shape,
                                       self.model.action_dist.param_shape)
            self.test_case.assertTrue(np.amax(step_action - np.array(action)) < 1e-4)
            self.test_case.assertTrue(abs(value - step_value) < 1e-4)

    def _truncate_first(self, rollouts):
        """
        Remove the first timestep on a random subset of
        rollouts.
        """
        do_trunc = np.random.choice(len(rollouts), len(rollouts)//2)
        for idx in do_trunc:
            # The second state isn't stored anywhere.
            rollout = rollouts[idx]
            first_obs = rollout.observations[0]
            outs = self.model.step([first_obs], rollout.start_state)

            rollout.start_state = outs['states']
            rollout.prev_steps = 1
            rollout.rewards = rollout.rewards[1:]
            rollout.model_outs = rollout.model_outs[1:]
            rollout.observations = rollout.observations[1:]
            rollout.infos = rollout.infos[1:]
        return rollouts

def _state_shape_uid(states):
    """
    Get an object that uniquely identifies the shape of
    the model state batch.
    """
    if states is None:
        return 'None'
    elif isinstance(states, tuple):
        return [_state_shape_uid(s) for s in states]
    return np.array(states).shape

class ACTest(unittest.TestCase):
    """
    Test actor-critic models.
    """
    def __init__(self, *args, **kwargs):
        super(ACTest, self).__init__(*args, **kwargs)
        self.session = tf.Session()
        env = TupleCartPole()
        try:
            action_space = env.action_space
            observation_space = env.observation_space
        finally:
            env.close()
        self.action_dist = gym_space_distribution(action_space)
        self.obs_vectorizer = gym_space_vectorizer(observation_space)

    def test_mlp(self):
        """
        Test the MLP model.
        """
        maker = lambda sess: MLP(sess, self.action_dist, self.obs_vectorizer,
                                 layer_sizes=[32])
        ModelTester(self, maker).test_all()

    def test_lstm(self):
        """
        Test an LSTM model.
        """
        maker = lambda sess: RNNCellAC(sess,
                                       self.action_dist,
                                       self.obs_vectorizer,
                                       make_cell=lambda: LSTMCell(32))
        ModelTester(self, maker).test_all()

    def test_multi_rnn(self):
        """
        Test a stacked LSTM with nested tuple state.
        """
        make_cell = lambda: MultiRNNCell([LSTMCell(16), LSTMCell(32)])
        maker = lambda sess: RNNCellAC(sess,
                                       self.action_dist,
                                       self.obs_vectorizer,
                                       make_cell=make_cell)
        ModelTester(self, maker).test_all()

if __name__ == '__main__':
    unittest.main()



"""
Logs for environment rollouts.
"""

import os
import time

import gym
import pandas

class LoggedEnv(gym.Wrapper):
    """
    An environment that logs episodes to a file.

    Logs are CSV files with three columns:
      r: episode reward
      l: episode length (timesteps)
      t: timestamp of episode end, relative to log start.
    """
    def __init__(self, env, log_path, use_locking=False):
        """
        Create a logged environment.

        Args:
          env: the environment to wrap.
          log_path: path to output CSV file.
          use_locking: use UNIX file locking to allow
            multiple processes to write to one log.

        If the log file already exists, it is appended to.
        """
        super(LoggedEnv, self).__init__(env)
        self._start_time = time.time()
        self._columns = ['r', 'l', 't']
        self._use_locking = use_locking
        self._file_desc = os.open(log_path, os.O_RDWR | os.O_CREAT)
        self._file = os.fdopen(self._file_desc, 'r+t')
        try:
            self._initialize_file()
        except:
            self._file.close()
            raise
        self._cur_timesteps = 0
        self._cur_reward = 0

    def _step(self, action):
        obs, rew, done, info = self.env.step(action)
        self._cur_reward += rew
        self._cur_timesteps += 1
        if done:
            self._write_entry()
            self._cur_timesteps = 0
            self._cur_reward = 0
        return obs, rew, done, info

    def _close(self):
        self.env.close()
        self._file.close()

    def _set_locked(self, locked):
        if not self._use_locking:
            return
        import fcntl
        if locked:
            fcntl.flock(self._file_desc, fcntl.LOCK_EX)
        else:
            fcntl.flock(self._file_desc, fcntl.LOCK_UN)

    def _write_entry(self):
        """
        Write the latest episode record.
        """
        info = {
            'r': [self._cur_reward],
            'l': [self._cur_timesteps],
            't': [time.time() - self._start_time]
        }
        data = pandas.DataFrame(info).to_csv(columns=self._columns, header=False, index=False)
        self._set_locked(True)
        try:
            self._file.seek(0, os.SEEK_END)
            self._file.write(data)
            self._file.flush()
        finally:
            self._set_locked(False)

    def _initialize_file(self):
        """
        Setup the initial file or read existing meta-data
        from it.
        """
        self._set_locked(True)
        self._file.seek(0, os.SEEK_END)
        if self._file.tell() > 0:
            self._file.seek(0, os.SEEK_SET)
            contents = pandas.read_csv(self._file)
            # pylint: disable=C1801
            if len(contents) > 0:
                self._start_time = time.time() - max(contents['t'])
        else:
            self._write_header()
        self._set_locked(False)

    def _write_header(self):
        """
        Write the CSV header.
        """
        self._file.write(','.join(self._columns) + '\n')
        self._file.flush()



"""
A synchronous version of advantage actor-critic.
"""

import tensorflow as tf

from .advantages import GAE
from . import util

# pylint: disable=R0902
class A2C:
    """
    Train TensorFlow actor-critic models using A2C.

    This works with any model that implements
    anyrl.models.TFActorCritic.

    Thanks to:
    https://github.com/openai/baselines/blob/699919f1cf2527b184f4445a3758a773f333a1ba/baselines/a2c/a2c.py
    """
    # pylint: disable=R0913
    def __init__(self,
                 model,
                 vf_coeff=0.5,
                 entropy_reg=0.01,
                 adv_est=GAE(lam=0.95, discount=0.99),
                 variables=None):
        self.model = model
        self.adv_est = adv_est

        self.variables = variables
        if variables is None:
            self.variables = tf.trainable_variables()

        self._advs = tf.placeholder(tf.float32, (None,))
        self._target_vals = tf.placeholder(tf.float32, (None,))
        self._actions = tf.placeholder(tf.float32,
                                       (None,) + model.action_dist.out_shape)

        self._create_objective(vf_coeff, entropy_reg)

    def feed_dict(self, rollouts, batch=None, advantages=None, targets=None):
        """
        Generate a TensorFlow feed_dict that feeds the
        rollouts into the objective.

        If no batch is specified, all rollouts are used.

        If advantages or targets are specified, then they
        are used instead of using the advantage estimator.
        """
        if batch is None:
            batch = next(self.model.batches(rollouts))
        advantages = advantages or self.adv_est.advantages(rollouts)
        targets = targets or self.adv_est.targets(rollouts)
        actions = util.select_model_out_from_batch('actions', rollouts, batch)
        feed_dict = batch['feed_dict']
        feed_dict[self._advs] = util.select_from_batch(advantages, batch)
        feed_dict[self._target_vals] = util.select_from_batch(targets, batch)
        feed_dict[self._actions] = self.model.action_dist.to_vecs(actions)
        return feed_dict

    def optimize(self,
                 max_grad_norm=0.5,
                 learning_rate=7e-4,
                 rms_decay=0.99,
                 rms_epsilon=1e-5):
        """
        Create an operation that trains the model based on
        values given by feed_dict.
        """
        grads = tf.gradients(tf.negative(self.objective), self.variables)
        if max_grad_norm is not None:
            grads, _ = tf.clip_by_global_norm(grads, max_grad_norm)
        trainer = tf.train.RMSPropOptimizer(learning_rate=learning_rate,
                                            decay=rms_decay,
                                            epsilon=rms_epsilon)
        return trainer.apply_gradients(list(zip(grads, self.variables)))

    def _create_objective(self, vf_coeff, entropy_reg):
        """
        Build up the objective graph.
        """
        actor, critic, mask = self.model.batch_outputs()
        dist = self.model.action_dist
        log_probs = dist.log_prob(actor, self._actions)
        entropies = dist.entropy(actor)
        critic_error = self._target_vals - critic
        self.actor_loss = -util.masked_mean(mask, log_probs * self._advs)
        self.critic_loss = util.masked_mean(mask, tf.square(critic_error))
        self.entropy = util.masked_mean(mask, entropies)
        self.objective = (entropy_reg * self.entropy - self.actor_loss -
                          vf_coeff * self.critic_loss)

    # TODO: API that supports schedules and runs the
    # entire training loop for us.



"""
Ways of gathering rollouts.
"""

from abc import ABC, abstractmethod
import time

from .rollout import empty_rollout

# pylint: disable=R0903
class Roller(ABC):
    """
    An object that gathers rollouts by running a model.
    """
    @abstractmethod
    def rollouts(self):
        """
        Return a list of Rollout objects.
        """
        pass

class BasicRoller(Roller):
    """
    Gathers episode rollouts from a Gym environment.
    """
    def __init__(self, env, model, min_episodes=1, min_steps=1):
        self.env = env
        self.model = model
        self.min_episodes = min_episodes
        self.min_steps = min_steps

    def rollouts(self):
        """
        Gather episodes until both self.min_episodes and
        self.min_steps are satisfied.
        """
        episodes = []
        num_steps = 0
        while num_steps < self.min_steps or len(episodes) < self.min_episodes:
            states = self.model.start_state(1)
            rollout = empty_rollout(states)
            obs = self.env.reset()
            while True:
                rollout.observations.append(obs)
                model_out = self.model.step([obs], states)
                rollout.model_outs.append(model_out)
                states = model_out['states']
                obs, rew, done, info = self.env.step(model_out['actions'][0])
                rollout.rewards.append(rew)
                rollout.infos.append(info)
                if done:
                    break
            num_steps += rollout.num_steps
            rollout.end_time = time.time()
            episodes.append(rollout)
        return episodes

class TruncatedRoller(Roller):
    """
    Gathers a fixed number of timesteps from each
    environment in a BatchedEnv.
    """
    def __init__(self, batched_env, model, num_timesteps):
        self.batched_env = batched_env
        self.model = model
        self.num_timesteps = num_timesteps

        # These end up being batches of sub-batches.
        # Each sub-batch corresponds to a sub-batch of
        # environments.
        self._last_states = None
        self._last_obs = None
        self._prev_steps = None
        self._prev_reward = None

    def reset(self):
        """
        Reset the environments, model states, and partial
        trajectory information.

        This needn't be called on new TruncatedRollers.
        """
        inner_dim = self.batched_env.num_envs_per_sub_batch
        outer_dim = self.batched_env.num_sub_batches
        self._last_obs = []
        self._prev_steps = [[0]*inner_dim for _ in range(outer_dim)]
        self._prev_reward = [[0]*inner_dim for _ in range(outer_dim)]
        for i in range(outer_dim):
            self.batched_env.reset_start(sub_batch=i)
        for i in range(outer_dim):
            self._last_obs.append(self.batched_env.reset_wait(sub_batch=i))
        self._last_states = [self.model.start_state(inner_dim)
                             for _ in range(outer_dim)]

    def rollouts(self):
        """
        Gather (possibly truncated) rollouts.
        """
        if self._last_states is None:
            self.reset()
        completed_rollouts = []
        running_rollouts = self._starting_rollouts()
        for _ in range(self.num_timesteps):
            self._step(completed_rollouts, running_rollouts)
        self._step(completed_rollouts, running_rollouts, final_step=True)
        self._add_truncated(completed_rollouts, running_rollouts)
        return completed_rollouts

    def _starting_rollouts(self):
        """
        Create empty rollouts with the start states and
        initial observations.
        """
        rollouts = []
        for batch_idx, states in enumerate(self._last_states):
            rollout_batch = []
            for env_idx in range(self.batched_env.num_envs_per_sub_batch):
                sub_state = _reduce_states(states, env_idx)
                prev_steps = self._prev_steps[batch_idx][env_idx]
                prev_reward = self._prev_reward[batch_idx][env_idx]
                rollout = empty_rollout(sub_state,
                                        prev_steps=prev_steps,
                                        prev_reward=prev_reward)
                rollout_batch.append(rollout)
            rollouts.append(rollout_batch)
        return rollouts

    # pylint: disable=R0914
    def _step(self, completed, running, final_step=False):
        """
        Wait for the previous batched step to complete (or
        use self._last_obs) and start a new step.

        Updates the running rollouts to reflect new steps
        and episodes.

        Returns the newest batch of model outputs.
        """
        for batch_idx, obses in enumerate(self._last_obs):
            if obses is None:
                step_out = self.batched_env.step_wait(sub_batch=batch_idx)
                obses, rews, dones, infos = step_out
                for env_idx, (rew, done, info) in enumerate(zip(rews, dones, infos)):
                    running[batch_idx][env_idx].rewards.append(rew)
                    running[batch_idx][env_idx].infos.append(info)
                    if done:
                        self._complete_rollout(completed, running, batch_idx, env_idx)
                    else:
                        self._prev_steps[batch_idx][env_idx] += 1
                        self._prev_reward[batch_idx][env_idx] += rew

            states = self._last_states[batch_idx]
            model_outs = self.model.step(obses, states)
            for env_idx, (obs, rollout) in enumerate(zip(obses, running[batch_idx])):
                reduced_out = _reduce_model_outs(model_outs, env_idx)
                rollout.observations.append(obs)
                rollout.model_outs.append(reduced_out)

            if final_step:
                self._last_obs[batch_idx] = obses
            else:
                self._last_states[batch_idx] = model_outs['states']
                self._last_obs[batch_idx] = None
                self.batched_env.step_start(model_outs['actions'], sub_batch=batch_idx)

    def _complete_rollout(self, completed, running, batch_idx, env_idx):
        """
        Finalize a rollout and start a new rollout.
        """
        running[batch_idx][env_idx].end_time = time.time()
        completed.append(running[batch_idx][env_idx])
        for prev in [self._prev_steps, self._prev_reward]:
            prev[batch_idx][env_idx] = 0
        start_state = self.model.start_state(1)
        _inject_state(self._last_states[batch_idx], start_state, env_idx)

        rollout = empty_rollout(start_state)
        running[batch_idx][env_idx] = rollout

    # pylint: disable=R0201
    def _add_truncated(self, completed, running):
        """
        Add partial but non-empty rollouts to completed.
        """
        for sub_running in running:
            for rollout in sub_running:
                if rollout.num_steps > 0:
                    rollout.end_time = time.time()
                    completed.append(rollout)

class EpisodeRoller(TruncatedRoller):
    """
    Gather rollouts from a BatchedEnv with step and
    episode quotas.

    An EpisodeRoller does not have any bias towards
    shorter episodes.
    As a result, it must gather at least as many episodes
    as there are environments in batched_env.
    """
    def __init__(self, batched_env, model, min_episodes=1, min_steps=1):
        self.min_episodes = min_episodes
        self.min_steps = min_steps

        # A batch of booleans in the shape of the envs.
        # An environment gets masked out once we have met
        # all the criteria and aren't looking for another
        # episode.
        self._env_mask = None

        super(EpisodeRoller, self).__init__(batched_env, model, 0)

    def reset(self):
        super(EpisodeRoller, self).reset()
        inner_dim = self.batched_env.num_envs_per_sub_batch
        outer_dim = self.batched_env.num_sub_batches
        self._env_mask = [[True]*inner_dim for _ in range(outer_dim)]

    def rollouts(self):
        """
        Gather full-episode rollouts.
        """
        self.reset()
        completed_rollouts = []
        running_rollouts = self._starting_rollouts()
        while self._any_envs_running():
            self._step(completed_rollouts, running_rollouts)
        # Make sure we are ready for the next reset().
        for batch_idx in range(self.batched_env.num_sub_batches):
            self.batched_env.step_wait(sub_batch=batch_idx)
        return completed_rollouts

    def _complete_rollout(self, completed, running, batch_idx, env_idx):
        comp_dest = completed
        if not self._env_mask[batch_idx][env_idx]:
            comp_dest = []
        super(EpisodeRoller, self)._complete_rollout(comp_dest, running,
                                                     batch_idx, env_idx)
        if self._criteria_met(completed):
            self._env_mask[batch_idx][env_idx] = False

    def _criteria_met(self, completed):
        """
        Check if the stopping criteria are met.
        """
        total_steps = sum([r.num_steps for r in completed])
        total_eps = len(completed)
        return total_steps >= self.min_steps and total_eps >= self.min_episodes

    def _any_envs_running(self):
        """
        Check if any environment is not masked out.
        """
        return any([any(masks) for masks in self._env_mask])

def _reduce_states(state_batch, env_idx):
    """
    Reduce a batch of states to a batch of one state.
    """
    if state_batch is None:
        return None
    elif isinstance(state_batch, tuple):
        return tuple(_reduce_states(s, env_idx) for s in state_batch)
    return state_batch[env_idx : env_idx+1]

def _inject_state(state_batch, state, env_idx):
    """
    Replace the state at the given index with a new state.
    """
    if state_batch is None:
        return
    elif isinstance(state_batch, tuple):
        return tuple(_inject_state(sb, s, env_idx)
                     for sb, s in zip(state_batch, state))
    state_batch[env_idx : env_idx+1] = state

def _reduce_model_outs(model_outs, env_idx):
    """
    Reduce a batch of model outputs to a batch of one
    model output.
    """
    out = dict()
    for key in model_outs:
        val = model_outs[key]
        if val is None:
            out[key] = None
        elif isinstance(val, tuple):
            out[key] = _reduce_states(val, env_idx)
        else:
            out[key] = val[env_idx : env_idx+1]
    return out



"""
Schedules for learning rates and other hyper-parameters.
"""

from abc import ABC, abstractmethod

import tensorflow as tf

class TFSchedule(ABC):
    """
    An abstract scheduled value for TensorFlow graphs.

    This can be used for annealed learning rates, among
    other things.

    Attributes:
      value: a Tensor with the current scheduled value.
      time: a Tensor indicating the current time.
      add_ph: a placeholder indicating how much time to
        add to the time counter.
      add_op: an Op which adds add_ph to time. It will
        never execute before self.value in the graph.
    """
    def __init__(self, dtype=tf.float32):
        time = tf.Variable(0, dtype=dtype, name='ScheduleCounter', trainable=False)
        self.value = self.compute_schedule(time)
        self.time = time
        self.add_ph = tf.placeholder(dtype)
        with tf.control_dependencies([self.value]):
            self.add_op = tf.assign_add(self.time, self.add_ph)

    def add_time(self, sess, amount):
        """
        Add the amount of time to the counter.

        Args:
          sess: the TensorFlow session.
          amount: the time to add.
        """
        sess.run(self.add_op, {self.add_ph: amount})

    @abstractmethod
    def compute_schedule(self, cur_time):
        """
        Compute the schedule value given the timestamp
        stored in cur_time.
        """
        pass

class LinearTFSchedule(TFSchedule):
    """
    A schedule that linearly interpolates between a start
    and an end value.
    """
    def __init__(self, duration=1.0, start_value=1.0, end_value=0.0, dtype=tf.float64):
        """
        Create a linear schedule.

        Args:
          duration: the timestamp at which the value
            should arrive at end_value.
          start_value: the initial value.
          end_value: the final value.
        """
        self._duration = float(duration)
        self._start_value = float(start_value)
        self._end_value = float(end_value)
        super(LinearTFSchedule, self).__init__(dtype=dtype)

    def compute_schedule(self, cur_time):
        frac_done = tf.clip_by_value(cur_time/self._duration, 0, 1)
        return (1-frac_done)*self._start_value + frac_done*self._end_value



"""
Wrappers for image preprocessing.
"""

import gym
import numpy as np

class DownsampleEnv(gym.ObservationWrapper):
    """
    An environment that downsamples its image inputs.
    """
    def __init__(self, env, rate):
        """
        Create a downsampling wrapper.

        Args:
          env: the environment to wrap.
          rate: (int) the downsample rate.
            New sizes are floor(old_size/rate).
        """
        super(DownsampleEnv, self).__init__(env)
        self._rate = rate
        old_space = env.observation_space
        self.observation_space = gym.spaces.Box(self._observation(old_space.low),
                                                self._observation(old_space.high))

    def _observation(self, observation):
        observation = observation[:1 + ((observation.shape[0] - 1) // self._rate) * self._rate,
                                  :1 + ((observation.shape[1] - 1) // self._rate) * self._rate]
        return observation[::self._rate, ::self._rate]

class GrayscaleEnv(gym.ObservationWrapper):
    """
    An environment that turns RGB images into grayscale.
    """
    def __init__(self, env, keep_depth=True, integers=True):
        """
        Create a grayscaling wrapper.

        Args:
          env: the environment to wrap.
          keep_depth: if True, a depth dimension is kept.
            Otherwise, the output is 2-D.
          integers: if True, the pixels are in [0, 255].
            Otherwise, they are in [0.0, 1.0].
        """
        super(GrayscaleEnv, self).__init__(env)
        old_space = env.observation_space
        self._integers = integers
        self._keep_depth = keep_depth
        self.observation_space = gym.spaces.Box(self._observation(old_space.low),
                                                self._observation(old_space.high))

    def _observation(self, observation):
        if self._integers:
            observation = observation // 3
        else:
            observation = observation / 3
        return np.sum(observation, axis=-1, keepdims=self._keep_depth, dtype=observation.dtype)

class FrameStackEnv(gym.Wrapper):
    """
    An environment that stacks images.

    Input arrays are stacked along the inner dimension.

    The stacking is ordered from oldest to newest.

    At the beginning of an episode, the first observation
    is repeated in order to complete the stack.
    """
    def __init__(self, env, num_images=2):
        """
        Create a frame stacking environment.

        Args:
          env: the environment to wrap.
          num_images: the number of images to stack.
            This includes the current observation.
        """
        super(FrameStackEnv, self).__init__(env)
        old_space = env.observation_space
        self.observation_space = gym.spaces.Box(np.repeat(old_space.low, num_images, axis=-1),
                                                np.repeat(old_space.high, num_images, axis=-1))
        self._num_images = num_images
        self._history = []

    def _reset(self, **kwargs):
        obs = super(FrameStackEnv, self)._reset(**kwargs)
        self._history = [obs] * self._num_images
        return np.concatenate(self._history, axis=-1)

    def _step(self, action):
        obs, rew, done, info = super(FrameStackEnv, self)._step(action)
        self._history.append(obs)
        self._history = self._history[1:]
        return np.concatenate(self._history, axis=-1), rew, done, info

class MaxEnv(gym.Wrapper):
    """
    An environment that takes the component-wise maximum
    over the last few observations.
    """
    def __init__(self, env, num_images=2):
        super(MaxEnv, self).__init__(env)
        self._num_images = num_images
        self._history = []

    def _reset(self, **kwargs):
        obs = super(MaxEnv, self)._reset(**kwargs)
        self._history = [obs]
        return np.max(self._history, axis=0)

    def _step(self, action):
        obs, rew, done, info = super(MaxEnv, self)._step(action)
        self._history.append(obs)
        self._history = self._history[-self._num_images:]
        return np.max(self._history, axis=0), rew, done, info

class ResizeImageEnv(gym.ObservationWrapper):
    """
    An environment that resizes observation images.

    This lazily imports TensorFlow when initialized.
    """
    def __init__(self, env, size=(84, 84), method=None):
        """
        Create a resizing wrapper.

        Args:
          env: the environment to wrap.
          size: the new (height, width) tuple.
          method: an interpolation method. Defaults to
            tf.image.ResizeMethod.AREA.
        """
        super(ResizeImageEnv, self).__init__(env)
        import tensorflow as tf
        config = tf.ConfigProto(device_count={'GPU': 0})
        self._sess = tf.Session(config=config)
        self._ph = tf.placeholder(tf.int32, shape=env.observation_space.shape)
        if method is None:
            method = tf.image.ResizeMethod.AREA
        self._resized = tf.image.resize_images(self._ph, size, method=method)
        old_space = env.observation_space
        self.observation_space = gym.spaces.Box(self._observation(old_space.low),
                                                self._observation(old_space.high))

    def _observation(self, observation):
        return self._sess.run(self._resized, feed_dict={self._ph: observation})



"""
Test various Roller implementations.
"""

import unittest

from anyrl.envs import batched_gym_env
from anyrl.rollouts import BasicRoller, TruncatedRoller, EpisodeRoller, Rollout
from anyrl.tests import SimpleEnv, SimpleModel
import numpy as np

class TruncatedRollerTest(unittest.TestCase):
    """
    Tests for TruncatedRoller.
    """
    def test_basic_equivalence(self):
        """
        Test that TruncatedRoller is equivalent to a
        BasicRoller when used with a single environment.
        """
        self._test_basic_equivalence_case(False, False)
        self._test_basic_equivalence_case(True, False)
        self._test_basic_equivalence_case(True, True)

    def test_truncation(self):
        """
        Test that sequence truncation works correctly for
        a batch of one environment.
        """
        self._test_truncation_case(False, False)
        self._test_truncation_case(True, False)
        self._test_truncation_case(True, True)

    def test_batch_equivalence(self):
        """
        Test that doing things in batches is equivalent to
        doing things one at a time.
        """
        self._test_batch_equivalence_case(False, False)
        self._test_batch_equivalence_case(True, False)
        self._test_batch_equivalence_case(True, True)

    def _test_basic_equivalence_case(self, stateful, state_tuple):
        """
        Test BasicRoller equivalence for a specific set of
        model settings.
        """
        env_fn = lambda: SimpleEnv(3, (4, 5), 'uint8')
        env = env_fn()
        model = SimpleModel(env.action_space.low.shape,
                            stateful=stateful,
                            state_tuple=state_tuple)
        basic_roller = BasicRoller(env, model, min_episodes=5)
        expected = basic_roller.rollouts()
        total_timesteps = sum([x.num_steps for x in expected])

        batched_env = batched_gym_env([env_fn], sync=True)
        trunc_roller = TruncatedRoller(batched_env, model, total_timesteps)
        actual = trunc_roller.rollouts()
        _compare_rollout_batch(self, actual, expected)

    def _test_truncation_case(self, stateful, state_tuple):
        """
        Test rollout truncation and continuation for a
        specific set of model parameters.
        """
        env_fn = lambda: SimpleEnv(7, (5, 3), 'uint8')
        env = env_fn()
        model = SimpleModel(env.action_space.low.shape,
                            stateful=stateful,
                            state_tuple=state_tuple)
        basic_roller = BasicRoller(env, model, min_episodes=5)
        expected = basic_roller.rollouts()
        total_timesteps = sum([x.num_steps for x in expected])

        batched_env = batched_gym_env([env_fn], sync=True)
        trunc_roller = TruncatedRoller(batched_env, model, total_timesteps//2 + 1)
        actual1 = trunc_roller.rollouts()
        self.assertTrue(actual1[-1].trunc_end)
        actual2 = trunc_roller.rollouts()
        expected1, expected2 = _artificial_truncation(expected,
                                                      len(actual1) - 1,
                                                      actual1[-1].num_steps)
        self.assertEqual(len(actual2), len(expected2)+1)
        actual2 = actual2[:-1]
        _compare_rollout_batch(self, actual1, expected1)
        _compare_rollout_batch(self, actual2, expected2)

    def _test_batch_equivalence_case(self, stateful, state_tuple):
        """
        Test that doing things in batches is consistent,
        given the model parameters.
        """
        env_fns = [lambda seed=x: SimpleEnv(seed, (5, 3), 'uint8') for x in range(15)]
        model = SimpleModel((5, 3),
                            stateful=stateful,
                            state_tuple=state_tuple)

        unbatched_rollouts = []
        for env_fn in env_fns:
            batched_env = batched_gym_env([env_fn], sync=True)
            trunc_roller = TruncatedRoller(batched_env, model, 17)
            for _ in range(3):
                unbatched_rollouts.extend(trunc_roller.rollouts())

        batched_rollouts = []
        batched_env = batched_gym_env(env_fns, num_sub_batches=3, sync=True)
        trunc_roller = TruncatedRoller(batched_env, model, 17)
        for _ in range(3):
            batched_rollouts.extend(trunc_roller.rollouts())

        _compare_rollout_batch(self, unbatched_rollouts, batched_rollouts,
                               ordered=False)

class EpisodeRollerTest(unittest.TestCase):
    """
    Tests for EpisodeRoller.
    """
    def test_basic_equivalence(self):
        """
        Test that EpisodeRoller is equivalent to a
        BasicRoller when run on a single environment.
        """
        state_opts = [(False, False), (True, False), (True, True)]
        kwargs_opts = [{'min_episodes': 5}, {'min_steps': 7}]
        for state_opt in state_opts:
            for kwargs_opt in kwargs_opts:
                self._test_basic_equivalence_case(*state_opt, **kwargs_opt)

    def test_batch_equivalence(self):
        """
        Test that EpisodeRoller is equivalent to a
        BasicRoller when run on a batch of envs.
        """
        state_opts = [(False, False), (True, False), (True, True)]
        kwargs_opts = [{'min_episodes': 5}, {'min_steps': 7}]
        for state_opt in state_opts:
            for kwargs_opt in kwargs_opts:
                self._test_batch_equivalence_case(*state_opt, **kwargs_opt)

    def test_multiple_batches(self):
        """
        Make sure calling rollouts multiple times works.
        """
        env_fn = lambda: SimpleEnv(3, (4, 5), 'uint8')
        env = env_fn()
        try:
            model = SimpleModel(env.action_space.low.shape)
        finally:
            env.close()
        batched_env = batched_gym_env([env_fn], sync=True)
        try:
            ep_roller = EpisodeRoller(batched_env, model, min_episodes=5, min_steps=7)
            first = ep_roller.rollouts()
            for _ in range(3):
                _compare_rollout_batch(self, first, ep_roller.rollouts())
        finally:
            batched_env.close()

    def _test_basic_equivalence_case(self, stateful, state_tuple,
                                     **roller_kwargs):
        """
        Test BasicRoller equivalence for a single env in a
        specific case.
        """
        env_fn = lambda: SimpleEnv(3, (4, 5), 'uint8')
        env = env_fn()
        model = SimpleModel(env.action_space.low.shape,
                            stateful=stateful,
                            state_tuple=state_tuple)
        basic_roller = BasicRoller(env, model, **roller_kwargs)
        expected = basic_roller.rollouts()

        batched_env = batched_gym_env([env_fn], sync=True)
        ep_roller = EpisodeRoller(batched_env, model, **roller_kwargs)
        actual = ep_roller.rollouts()
        _compare_rollout_batch(self, actual, expected)

    def _test_batch_equivalence_case(self, stateful, state_tuple,
                                     **roller_kwargs):
        """
        Test BasicRoller equivalence when using a batch of
        environments.
        """
        env_fn = lambda: SimpleEnv(3, (4, 5), 'uint8')
        model = SimpleModel((4, 5),
                            stateful=stateful,
                            state_tuple=state_tuple)

        batched_env = batched_gym_env([env_fn]*21, num_sub_batches=7, sync=True)
        ep_roller = EpisodeRoller(batched_env, model, **roller_kwargs)
        actual = ep_roller.rollouts()

        total_steps = sum([r.num_steps for r in actual])
        self.assertTrue(len(actual) >= ep_roller.min_episodes)
        self.assertTrue(total_steps >= ep_roller.min_steps)

        if 'min_steps' not in roller_kwargs:
            num_eps = ep_roller.min_episodes + batched_env.num_envs - 1
            self.assertTrue(len(actual) == num_eps)

        basic_roller = BasicRoller(env_fn(), model, min_episodes=len(actual))
        expected = basic_roller.rollouts()

        _compare_rollout_batch(self, actual, expected)

def _compare_rollout_batch(test, rs1, rs2, ordered=True):
    """
    Assert that batches of rollouts are the same.
    """
    test.assertEqual(len(rs1), len(rs2))
    if ordered:
        for rollout1, rollout2 in zip(rs1, rs2):
            test.assertEqual(_rollout_hash(rollout1), _rollout_hash(rollout2))
    else:
        hashes1 = [_rollout_hash(r) for r in rs1]
        hashes2 = [_rollout_hash(r) for r in rs2]
        for hash1 in hashes1:
            test.assertTrue(hash1 in hashes2)
            hashes2.remove(hash1)

def _rollout_hash(rollout):
    """
    Generate a string that uniquely identifies a rollout.
    """
    # Prevent ellipsis.
    np.set_printoptions(threshold=1e6)

    res = ''
    res += str(rollout.trunc_start)
    res += str(rollout.trunc_end)
    res += str(rollout.prev_steps)
    res += '|'
    res += str(rollout.prev_reward)
    res += str(rollout.observations)
    res += str(rollout.rewards)
    res += str(rollout.infos)
    for out in rollout.model_outs:
        res += 'output'
        for key in sorted(out.keys()):
            res += 'kv'
            res += key
            res += str(out[key])
    return res

def _artificial_truncation(rollouts, rollout_idx, timestep_idx):
    """
    Split up the rollouts into two batches by artificially
    truncating (and splitting) the given rollout before
    the given timestep.

    Returns (left_rollouts, right_rollouts)
    """
    to_split = rollouts[rollout_idx]
    left = Rollout(observations=to_split.observations[:timestep_idx+1],
                   model_outs=to_split.model_outs[:timestep_idx+1],
                   rewards=to_split.rewards[:timestep_idx],
                   start_state=to_split.start_state,
                   infos=to_split.infos[:timestep_idx])
    right = Rollout(observations=to_split.observations[timestep_idx:],
                    model_outs=to_split.model_outs[timestep_idx:],
                    rewards=to_split.rewards[timestep_idx:],
                    start_state=to_split.model_outs[timestep_idx-1]['states'],
                    prev_steps=timestep_idx,
                    prev_reward=left.total_reward,
                    infos=to_split.infos[timestep_idx:])
    return rollouts[:rollout_idx]+[left], [right]+rollouts[rollout_idx+1:]

if __name__ == '__main__':
    unittest.main()



"""
Abstractions for probability distributions.
"""

from abc import ABC, abstractmethod, abstractproperty

class Vectorizer(ABC):
    """
    A way to convert action or observation space elements
    to tensors.
    """
    @abstractproperty
    def out_shape(self):
        """
        The shape of vectorized space elements.
        """
        pass

    @abstractmethod
    def to_vecs(self, space_elements):
        """
        Convert a list-like object of space elements to a
        list-like object of tensors.
        """
        pass

class Distribution(Vectorizer):
    """
    A parametric probability distribution.

    All methods operate on and produce TensorFlow tensors
    except for sample().
    """
    @abstractproperty
    def param_shape(self):
        """
        Get the shape of the distribution parameters.
        """
        pass

    @abstractmethod
    def sample(self, param_batch):
        """
        Create a list of samples from the distribution
        given the batch of parameter vectors.

        The param_batch should be some kind of array-like
        object.
        The result is a list-like object of space
        elements.
        """
        pass

    @abstractmethod
    def mode(self, param_batch):
        """
        Compute the most likely sample for each parameter
        vector in a batch of parameter vectors.

        The param_batch should be some kind of array-like
        object.
        The result is a list-like object of space
        elements.
        """
        pass

    @abstractmethod
    def log_prob(self, param_batch, sample_vecs):
        """
        Compute the log probability (or log density) of
        the samples, given the parameters.

        You can obtain vectors for sample_vecs via the
        to_vecs() method.
        """
        pass

    @abstractmethod
    def entropy(self, param_batch):
        """
        Compute the entropy (or differential entropy) for
        each set of parameters.
        """
        pass

    @abstractmethod
    def kl_divergence(self, param_batch_1, param_batch_2):
        """
        Compute KL(params_1 || params_2) for each pair of
        parameters in the batch.
        """
        pass



"""
Tests for advantage routines.
"""

import unittest

from anyrl.rollouts import Rollout
from anyrl.algos.advantages import GAE

class TestEstimators(unittest.TestCase):
    """
    Test AdvantageEstimator implementations.
    """
    def test_gae(self):
        """
        Test generalized advantage estimation.
        """
        rollouts = [
            _dummy_rollout([1, 0.5, 3], [0, -1.5, 2]),
            _dummy_rollout([3, 2, -7], [1, 9])
        ]
        judger = GAE(lam=0.7, discount=0.9)
        actual = judger.advantages(rollouts)
        expected = [[-0.5059, 0.07, -1], [0.241, 0.7]]
        for actual_seq, expected_seq in zip(actual, expected):
            self.assertEqual(len(actual_seq), len(expected_seq))
            for act, exp in zip(actual_seq, expected_seq):
                self.assertTrue(abs(act-exp) < 1e-5)

def _dummy_rollout(pred_vals, rewards):
    """
    Create a rollout with the predicted values and actual
    rewards.
    """
    model_outs = [{'values': [x]} for x in pred_vals]
    return Rollout(observations=[[1]]*len(pred_vals),
                   model_outs=model_outs,
                   rewards=rewards,
                   start_state=None)

if __name__ == '__main__':
    unittest.main()



"""
Ways of running gym environments.
"""

from abc import ABC, abstractmethod, abstractproperty

class AsyncEnv(ABC):
    """
    An asynchronous environment.
    """
    @abstractmethod
    def reset_start(self):
        """
        Start resetting the environment.

        This should not be called while any other
        asynchronous operations are taking place.
        """
        pass

    @abstractmethod
    def reset_wait(self):
        """
        Wait for a reset_start() to finish.

        Returns an observation.

        The resulting array belongs to the caller.
        It should not be modified after-the-fact by the
        environment.
        """
        pass

    @abstractmethod
    def step_start(self, action):
        """
        Start taking a step in the environment.

        This should not be called while any other
        asynchronous operations are taking place.
        """
        pass

    @abstractmethod
    def step_wait(self):
        """
        Wait for a step_start() to finish.

        Returns (observation, reward, done, info).

        If done is true, then the environment was reset
        and the new observation was taken.

        The resulting arrays belong to the caller.
        They should not be modified after-the-fact by the
        environment.
        """
        pass

    @abstractmethod
    def close(self):
        """
        Clean up the environment's resources.

        Waits for any pending operations to complete.
        """
        pass

class BatchedEnv(ABC):
    """
    A set of environments running in batches.

    The batch of environments is divided up into
    equally sized sub-batches.
    Each sub-batch is a set of environments that run in
    lockstep.

    Different BatchedEnvs may schedule jobs in different
    ways, but generally it will be FIFO order.
    Thus, it's best to wait for jobs in the same order
    that you started them.
    """
    @property
    def num_envs(self):
        """
        The total number of environments.
        """
        return self.num_sub_batches * self.num_envs_per_sub_batch

    @abstractproperty
    def num_sub_batches(self):
        """
        The number of sub-batches.
        """
        pass

    @abstractproperty
    def num_envs_per_sub_batch(self):
        """
        The number of environments per sub-batch.
        """
        pass

    @abstractmethod
    def reset_start(self, sub_batch=0):
        """
        Start resetting the sub-batch.

        This should not be called while any other
        operations are taking place for the sub-batch.
        """
        pass

    @abstractmethod
    def reset_wait(self, sub_batch=0):
        """
        Wait for a reset_start() to finish.

        Returns a list-like object of observations.

        The resulting array belongs to the caller.
        It should not be modified after-the-fact by the
        environment.
        """
        pass

    @abstractmethod
    def step_start(self, actions, sub_batch=0):
        """
        Start taking a step in the batch of environments.
        Takes a list-like object of actions.

        This should not be called while any other
        asynchronous operations are taking place.
        """
        pass

    @abstractmethod
    def step_wait(self, sub_batch=0):
        """
        Wait for a step_start() to finish.

        Returns (observations, rewards, dones, infos),
        where all those are list-like objects.

        If a done value is true, then the environment was
        automatically reset and the new observation was
        returned.

        The resulting arrays belong to the caller.
        They should not be modified after-the-fact by the
        environment.
        """
        pass

    @abstractmethod
    def close(self):
        """
        Clean up the environments' resources.

        Waits for any pending operations to complete.
        """
        pass

class BatchedAsyncEnv(BatchedEnv):
    """
    A BatchedEnv that controls AsyncEnvs.

    If the first AsyncEnv has an action_space and/or
    observation_space attribute, those attributes are
    copied.
    """
    def __init__(self, sub_batches):
        # pylint: disable=C1801
        assert len(sub_batches) > 0
        first_len = len(sub_batches[0])
        assert all([len(x) == first_len for x in sub_batches])
        self._sub_batches = sub_batches

        self.action_space = None
        self.observation_space = None

        first_env = sub_batches[0][0]
        if hasattr(first_env, 'action_space'):
            self.action_space = first_env.action_space
        if hasattr(first_env, 'observation_space'):
            self.observation_space = first_env.observation_space

    @property
    def num_sub_batches(self):
        return len(self._sub_batches)

    @property
    def num_envs_per_sub_batch(self):
        return len(self._sub_batches[0])

    def reset_start(self, sub_batch=0):
        for env in self._sub_batches[sub_batch]:
            env.reset_start()

    def reset_wait(self, sub_batch=0):
        obses = []
        for env in self._sub_batches[sub_batch]:
            obses.append(env.reset_wait())
        return obses

    def step_start(self, actions, sub_batch=0):
        assert len(actions) == self.num_envs_per_sub_batch
        for env, action in zip(self._sub_batches[sub_batch], actions):
            env.step_start(action)

    def step_wait(self, sub_batch=0):
        obses, rews, dones, infos = ([], [], [], [])
        for env in self._sub_batches[sub_batch]:
            obs, rew, done, info = env.step_wait()
            obses.append(obs)
            rews.append(rew)
            dones.append(done)
            infos.append(info)
        return obses, rews, dones, infos

    def close(self):
        for batch in self._sub_batches:
            for env in batch:
                env.close()



"""
Tests for environment wrappers.
"""

import os
import shutil
import tempfile
import unittest

import gym
import numpy as np
import pandas
import tensorflow as tf

from anyrl.envs import batched_gym_env
from anyrl.envs.wrappers import (RL2Env, DownsampleEnv, GrayscaleEnv, FrameStackEnv,
                                 MaxEnv, ResizeImageEnv, BatchedFrameStack, ObservationPadEnv,
                                 LoggedEnv)
from anyrl.tests import SimpleEnv

class RL2EnvTest(unittest.TestCase):
    """
    Tests for RL2Env.
    """
    def test_num_eps(self):
        """
        Test that the meta-episode contains the right
        number of sub-episodes.
        """
        real_env = SimpleEnv(1, (3, 4), 'uint8')
        env = RL2Env(real_env, real_env.action_space.sample(),
                     num_eps=3, warmup_eps=1)
        done_eps = 0
        env.reset()
        while done_eps < 3:
            obs, _, done, _ = env.step(env.action_space.sample())
            if obs[3]:
                done_eps += 1
            if done_eps == 3:
                self.assertTrue(done)
                break
            else:
                self.assertFalse(done)

    def test_rewards(self):
        """
        Test that rewards are masked properly.
        """
        real_env = SimpleEnv(1, (3, 4), 'uint8')
        env = RL2Env(real_env, real_env.action_space.sample(), num_eps=5, warmup_eps=-2)
        env.reset()
        done_eps = 0
        while done_eps < 3:
            obs, rew, _, _ = env.step(env.action_space.sample())
            self.assertEqual(rew, 0)
            if obs[3]:
                done_eps += 1
        while True:
            _, rew, done, _ = env.step(env.action_space.sample())
            self.assertNotEqual(rew, 0)
            if done:
                break

class DownsampleEnvTest(unittest.TestCase):
    """
    Tests for DownsampleEnv.
    """
    def test_rate_1(self):
        """
        Test with rate=1.
        """
        low = np.array([[1, 2], [3, 4]])
        high = np.array([[3, 4], [5, 6]])
        env = DownsampleEnv(ShapeEnv(low, high), 1)
        self.assertTrue((env.observation_space.low == low).all())
        self.assertTrue((env.observation_space.high == high).all())

    def test_rate_2(self):
        """
        Test with rate=2.
        """
        low = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        high = np.array([[9, 10, 11, 12], [13, 14, 15, 16]])
        env = DownsampleEnv(ShapeEnv(low, high), 2)
        self.assertEqual(env.observation_space.shape, (1, 2))
        self.assertTrue((env.observation_space.low == np.array([[1, 3]])).all())
        self.assertTrue((env.observation_space.high == np.array([[9, 11]])).all())

        low = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 6, 7, 8]])
        high = np.array([[9, 10, 11, 12], [13, 14, 15, 16], [16, 14, 15, 16]])
        env = DownsampleEnv(ShapeEnv(low, high), 2)
        self.assertEqual(env.observation_space.shape, (2, 2))
        self.assertTrue((env.observation_space.low == np.array([[1, 3], [9, 7]])).all())
        self.assertTrue((env.observation_space.high == np.array([[9, 11], [16, 15]])).all())

class GrayscaleEnvTest(unittest.TestCase):
    """
    Tests for GrayscaleEnv.
    """
    def test_integers(self):
        """
        Tests for integer observations.
        """
        low = np.array([[[0, 0, 0], [16, 16, 50]]])
        high = np.array([[[19, 255, 13], [255, 255, 255]]])
        env = GrayscaleEnv(ShapeEnv(low, high))
        self.assertEqual(env.observation_space.shape, (1, 2, 1))
        self.assertTrue((env.observation_space.low == np.array([[[0], [26]]])).all())
        self.assertTrue((env.observation_space.high == np.array([[[95], [255]]])).all())

class FrameStackEnvTest(unittest.TestCase):
    """
    Tests for FrameStackEnv.
    """
    def test_stack_3(self):
        """
        Test stacking 3 frames.
        """
        low = np.zeros((4, 5, 2))
        high = np.zeros((4, 5, 2)) + 0xff
        env = FrameStackEnv(ShapeEnv(low, high), 3)
        self.assertEqual(env.observation_space.shape, (4, 5, 6))
        obs1 = env.reset()
        obs2, _, _, _ = env.step(0)
        obs3, _, _, _ = env.step(0)
        self.assertFalse((obs1 == obs2).all())
        self.assertFalse((obs1 == obs3).all())
        self.assertFalse((obs2 == obs3).all())
        self.assertTrue((obs1[:, :, 2:] == obs2[:, :, :4]).all())
        self.assertTrue((obs2[:, :, 2:] == obs3[:, :, :4]).all())

class MaxEnvTest(unittest.TestCase):
    """
    Tests for MaxEnv.
    """
    def test_max_2(self):
        """
        Test maxing 2 frames.
        """
        env = SimpleEnv(5, (3, 2, 5), 'float32')
        actions = [env.action_space.sample() for _ in range(4)]
        frame1 = env.reset()
        frame2 = env.step(actions[0])[0]
        frame3 = env.step(actions[1])[0]
        frame4 = env.step(actions[2])[0]
        frame5 = env.step(actions[3])[0]
        wrapped = MaxEnv(env, num_images=2)
        max1 = wrapped.reset()
        max2 = wrapped.step(actions[0])[0]
        max3 = wrapped.step(actions[1])[0]
        max4 = wrapped.step(actions[2])[0]
        max5 = wrapped.step(actions[3])[0]
        self.assertTrue((max1 == frame1).all())
        self.assertTrue((max2 == np.max([frame1, frame2], axis=0)).all())
        self.assertTrue((max3 == np.max([frame2, frame3], axis=0)).all())
        self.assertTrue((max4 == np.max([frame3, frame4], axis=0)).all())
        self.assertTrue((max5 == np.max([frame4, frame5], axis=0)).all())

class ResizeImageEnvTest(unittest.TestCase):
    """
    Tests for ResizeImageEnv.
    """
    def test_resize_even(self):
        """
        Test resizing for an even number of pixels.
        """
        env = SimpleEnv(5, (13, 5, 3), 'float32')
        frame = env.reset()
        actual = ResizeImageEnv(env, size=(5, 4)).reset()
        expected = tf.Session().run(
            tf.image.resize_images(frame, [5, 4], method=tf.image.ResizeMethod.AREA))
        self.assertEqual(actual.shape, (5, 4, 3))
        self.assertTrue(np.allclose(actual, expected))

class BatchedFrameStackTest(unittest.TestCase):
    """
    Tests for BatchedFrameStack.
    """
    def test_equivalence(self):
        """
        Test that BatchedFrameStack is equivalent to a
        regular batched FrameStackEnv.
        """
        envs = [lambda idx=i: SimpleEnv(idx+2, (3, 2, 5), 'float32') for i in range(6)]
        env1 = BatchedFrameStack(batched_gym_env(envs, num_sub_batches=3, sync=True))
        env2 = batched_gym_env([lambda env=e: FrameStackEnv(env()) for e in envs],
                               num_sub_batches=3, sync=True)
        for j in range(50):
            for i in range(3):
                if j == 0 or (j + i) % 17 == 0:
                    env1.reset_start(sub_batch=i)
                    env2.reset_start(sub_batch=i)
                    obs1 = env1.reset_wait(sub_batch=i)
                    obs2 = env2.reset_wait(sub_batch=i)
                    self.assertTrue(np.allclose(obs1, obs2))
                actions = [env1.action_space.sample() for _ in range(2)]
                env1.step_start(actions, sub_batch=i)
                env2.step_start(actions, sub_batch=i)
                obs1, rews1, dones1, _ = env1.step_wait(sub_batch=i)
                obs2, rews2, dones2, _ = env2.step_wait(sub_batch=i)
                self.assertTrue(np.allclose(obs1, obs2))
                self.assertTrue(np.array(rews1 == rews2).all())
                self.assertTrue(np.array(dones1 == dones2).all())

class ObservationPadEnvTest(unittest.TestCase):
    """
    Tests for ObservationPadEnv.
    """
    def test_centered(self):
        """
        Test centered padding.
        """
        env = ShapeEnv(np.array([[1, 2, 5], [3, 4, 0]]),
                       np.array([[12, 21, 15], [31, 14, 10]]))
        padded = ObservationPadEnv(env, (3, 5))
        self.assertEqual(padded.observation_space.shape, (3, 5))
        self.assertTrue(np.allclose(padded.observation_space.low,
                                    np.array([[0, 1, 2, 5, 0], [0, 3, 4, 0, 0], [0]*5])))
        self.assertTrue(np.allclose(padded.observation_space.high,
                                    np.array([[0, 12, 21, 15, 0], [0, 31, 14, 10, 0], [0]*5])))

    def test_uncentered(self):
        """
        Test uncentered padding.
        """
        env = ShapeEnv(np.array([[1, 2, 5], [3, 4, 0]]),
                       np.array([[12, 21, 15], [31, 14, 10]]))
        padded = ObservationPadEnv(env, (3, 5), center=False)
        self.assertEqual(padded.observation_space.shape, (3, 5))
        self.assertTrue(np.allclose(padded.observation_space.low,
                                    np.array([[1, 2, 5, 0, 0], [3, 4, 0, 0, 0], [0]*5])))
        self.assertTrue(np.allclose(padded.observation_space.high,
                                    np.array([[12, 21, 15, 0, 0], [31, 14, 10, 0, 0], [0]*5])))

class LoggedEnvTest(unittest.TestCase):
    """
    Tests for LoggedEnv.
    """
    def test_single_env(self):
        """
        Test monitoring for a single environment.
        """
        dirpath = tempfile.mkdtemp()
        try:
            log_file = os.path.join(dirpath, 'monitor.csv')
            env = LoggedEnv(SimpleEnv(2, (3,), 'float32'), log_file)
            for _ in range(4):
                env.reset()
                while not env.step(env.action_space.sample())[2]:
                    pass
            env.close()
            with open(log_file, 'rt'):
                log_contents = pandas.read_csv(log_file)
                self.assertEqual(list(log_contents['r']), [2] * 4)
                self.assertEqual(list(log_contents['l']), [3] * 4)
        finally:
            shutil.rmtree(dirpath)

    def test_multi_env(self):
        """
        Test monitoring for concurrent environments.
        """
        dirpath = tempfile.mkdtemp()
        try:
            log_file = os.path.join(dirpath, 'monitor.csv')
            env1 = LoggedEnv(SimpleEnv(2, (3,), 'float32'), log_file, use_locking=True)
            env2 = LoggedEnv(SimpleEnv(3, (3,), 'float32'), log_file, use_locking=True)

            env1.reset()
            env2.reset()
            for _ in range(13):
                for env in [env1, env2]:
                    if env.step(env.action_space.sample())[2]:
                        env.reset()
            env1.close()
            env2.close()

            with open(log_file, 'rt'):
                log_contents = pandas.read_csv(log_file)
                self.assertEqual(list(log_contents['r']), [2, 2.5, 2, 2.5, 2, 2, 2.5])
                self.assertEqual(list(log_contents['l']), [3, 4, 3, 4, 3, 3, 4])
        finally:
            shutil.rmtree(dirpath)

class ShapeEnv(gym.Env):
    """
    An environment with a pre-defined observation shape.
    """
    def __init__(self, low, high):
        super(ShapeEnv, self).__init__()
        self.observation_space = gym.spaces.Box(low, high)

    def _reset(self):
        return self.observation_space.sample()

    def _step(self, action):
        return self.observation_space.sample(), 0, False, {}

if __name__ == '__main__':
    unittest.main()



"""
Tests for various environment APIs.
"""

import sys
import unittest

import numpy as np

from anyrl.envs import batched_gym_env, AsyncGymEnv
from anyrl.tests import SimpleEnv

SUB_BATCH_SIZE = 4
NUM_SUB_BATCHES = 3
NUM_STEPS = 100
SHAPE = (3, 8)

class BatchedGymEnvTest(unittest.TestCase):
    """
    Tests for the environment produced by batched_gym_env.
    """
    def test_dummy_equiv_bytes(self):
        """
        Test equivalence with uint8 observations.

        Intended to test shared memory.
        """
        self._test_dummy_equiv_dtype('uint8')

    def test_dummy_equiv_floats(self):
        """
        Test equivalence with float32 observations.
        """
        self._test_dummy_equiv_dtype('float32')

    def _test_dummy_equiv_dtype(self, dtype):
        """
        Test that batched_gym_env() gives something
        equivalent to a synchronous environment.
        """
        def make_fn(seed):
            """
            Get an environment constructor with a seed.
            """
            return lambda: SimpleEnv(seed, SHAPE, dtype)
        fns = [make_fn(i) for i in range(SUB_BATCH_SIZE * NUM_SUB_BATCHES)]
        real = batched_gym_env(fns, num_sub_batches=NUM_SUB_BATCHES)
        dummy = batched_gym_env(fns, num_sub_batches=NUM_SUB_BATCHES, sync=True)
        try:
            self._assert_resets_equal(dummy, real)
            np.random.seed(1337)
            for _ in range(NUM_STEPS):
                joint_shape = (SUB_BATCH_SIZE,) + SHAPE
                actions = np.array(np.random.randint(0, 0x100, size=joint_shape),
                                   dtype=dtype)
                self._assert_steps_equal(actions, dummy, real)
        finally:
            dummy.close()
            real.close()

    def _assert_resets_equal(self, env1, env2):
        # Non-overlapping resets.
        for i in range(env1.num_sub_batches):
            env1.reset_start(sub_batch=i)
            env2.reset_start(sub_batch=i)
            self._assert_reset_waits_equal(env1, env2, i)

        # Overlapping & async resets.
        for i in range(env1.num_sub_batches):
            env1.reset_start(sub_batch=i)
            env2.reset_start(sub_batch=i)
        for i in range(env1.num_sub_batches):
            self._assert_reset_waits_equal(env1, env2, i)

    def _assert_reset_waits_equal(self, env1, env2, sub_batch):
        out1 = np.array(env1.reset_wait(sub_batch=sub_batch))
        out2 = np.array(env2.reset_wait(sub_batch=sub_batch))
        self.assertTrue((out1 == out2).all())

    def _assert_steps_equal(self, actions, env1, env2):
        # Non-overlapping steps.
        for i in range(env1.num_sub_batches):
            env1.step_start(actions, sub_batch=i)
            env2.step_start(actions, sub_batch=i)
            self._assert_step_waits_equal(env1, env2, i)

        # Overlapping & async steps.
        for i in range(env1.num_sub_batches):
            env1.step_start(actions, sub_batch=i)
            env2.step_start(actions, sub_batch=i)
        for i in range(env1.num_sub_batches):
            self._assert_step_waits_equal(env1, env2, i)

    def _assert_step_waits_equal(self, env1, env2, sub_batch):
        outs1 = env1.step_wait(sub_batch=sub_batch)
        outs2 = env2.step_wait(sub_batch=sub_batch)
        for out1, out2 in zip(outs1[:3], outs2[:3]):
            self.assertTrue((np.array(out1) == np.array(out2)).all())
        self.assertEqual(outs1[3], outs2[3])

class EnvFailuresTest(unittest.TestCase):
    """
    Tests for AsyncEnv failure handling.
    """
    def test_exit(self):
        """
        Test an environment that straightup exits.
        """
        try:
            AsyncGymEnv(lambda: sys.exit(1), None)
        except RuntimeError:
            return
        self.fail('should have gotten exception')

    def test_exception(self):
        """
        Test an environment that throws.
        """
        try:
            def raiser(): # pylint: disable=C0111
                raise ValueError('hello world')
            AsyncGymEnv(raiser, None)
        except RuntimeError:
            return
        self.fail('should have gotten exception')

if __name__ == '__main__':
    unittest.main()



"""
APIs for aggregate spaces.
"""

import numpy as np
import tensorflow as tf

from .base import Distribution

class TupleDistribution(Distribution):
    """
    A distribution that consists of an ordered list of
    sub-distributions.
    """
    def __init__(self, sub_dists):
        self.tuple = tuple(sub_dists)

    @property
    def out_shape(self):
        return (int(sum([np.prod(d.out_shape) for d in self.tuple])),)

    def to_vecs(self, space_elements):
        per_dist = zip(*space_elements)
        sub_vecs = [d.to_vecs(list(l)) for d, l in zip(self.tuple, per_dist)]
        flat_shape = (len(space_elements), -1)
        flat_batches = [np.reshape(np.array(v), flat_shape) for v in sub_vecs]
        return np.concatenate(flat_batches, axis=-1)

    @property
    def param_shape(self):
        return (int(sum([np.prod(d.param_shape) for d in self.tuple])),)

    def sample(self, param_batch):
        per_dist = self.unpack_params(np.array(param_batch))
        samples = [d.sample(p) for d, p in zip(self.tuple, per_dist)]
        return list(zip(*samples))

    def mode(self, param_batch):
        per_dist = self.unpack_params(np.array(param_batch))
        modes = [d.mode(p) for d, p in zip(self.tuple, per_dist)]
        return list(zip(*modes))

    def log_prob(self, param_batch, sample_vecs):
        per_dist = self.unpack_params(param_batch)
        samples = self.unpack_outs(sample_vecs)
        sub_probs = [d.log_prob(p, s) for d, p, s
                     in zip(self.tuple, per_dist, samples)]
        return tf.add_n(sub_probs)

    def entropy(self, param_batch):
        per_dist = self.unpack_params(param_batch)
        sub_ents = [d.entropy(p) for d, p in zip(self.tuple, per_dist)]
        return tf.add_n(sub_ents)

    def kl_divergence(self, param_batch_1, param_batch_2):
        per_dist_1 = self.unpack_params(param_batch_1)
        per_dist_2 = self.unpack_params(param_batch_2)
        sub_kls = [d.kl_divergence(p1, p2) for d, p1, p2
                   in zip(self.tuple, per_dist_1, per_dist_2)]
        return tf.add_n(sub_kls)

    def unpack_outs(self, sample_vecs):
        """
        Unpack tensors from to_vecs into a tuple of
        properly-shaped tensors for each sub-space.

        Takes either a numpy array or a TF tensor.
        """
        return _unpack(sample_vecs, [d.out_shape for d in self.tuple])

    def unpack_params(self, param_batch):
        """
        Unpack parameter tensors into a tuple of
        properly-shaped tensors for each sub-space.

        Takes either a numpy array or a TF tensor.
        """
        return _unpack(param_batch, [d.param_shape for d in self.tuple])

def _unpack(vecs, shapes):
    """
    Unpack the flattened and packed vectors into a tuple.
    """
    offset = 0
    res = ()
    for shape in shapes:
        size = int(np.prod(shape))
        sub_slice = vecs[:, offset:offset+size]
        offset += size
        if isinstance(vecs, np.ndarray):
            sub_shape = (len(vecs),) + shape
            res += (np.reshape(sub_slice, sub_shape),)
        else:
            sub_shape = (tf.shape(vecs)[0],) + shape
            res += (tf.reshape(sub_slice, sub_shape),)
    return res



"""
Test reward normalizers.
"""

import unittest

import numpy as np

from anyrl.rollouts.norm import OnlineAverage

class TestOnlineAverage(unittest.TestCase):
    """
    Test online averages.
    """
    def test_running_average(self):
        """
        Test a running, unbiased average.
        """
        avg = OnlineAverage(rate=None)
        values = np.random.normal(size=15)
        for subset in [values[:3], values[3:9], values[9:14], values[14:]]:
            avg.update(subset)
        self.assertTrue(np.allclose(avg.value, np.mean(values)))

    def test_moving_average(self):
        """
        Test an exponential moving average.
        """
        avg = OnlineAverage(rate=0.05)
        values = np.random.normal(size=15)
        subsets = [values[:3], values[3:9], values[9:14], values[14:]]
        for subset in subsets:
            avg.update(subset)
        expected = np.mean(subsets[0])
        for subset in subsets[1:]:
            expected += 0.05 * (np.mean(subset) - expected)
        self.assertTrue(np.allclose(avg.value, expected))

if __name__ == '__main__':
    unittest.main()



"""
Stateful neural network models.
"""

import numpy as np
import tensorflow as tf
from tensorflow.contrib.framework import nest # pylint: disable=E0611
from tensorflow.contrib.layers import fully_connected # pylint: disable=E0611

from .base import TFActorCritic
from .util import mini_batches, product, mix_init_states

# pylint: disable=E1129

# pylint: disable=R0902
class RecurrentAC(TFActorCritic):
    """
    A base class for any stateful actor-critic model.

    Sets the following special attributes:
      seq_lens_ph: placeholder of sequence lengths
      is_init_state_ph: placeholder of booleans indicating
        if sequences start with an initial state.

    Subclasses should set several attributes on init:
      init_state_vars: variable or tuple of variables
        representing the start state vectors.
        This can be set automatically be calling
        create_state_fields().
      obs_ph: placeholder for input sequences
      first_state_ph: placeholder or tuple of placeholders
        for the states at the beginning of the sequences.
      mask_ph: placeholder for timestep mask. Should be of
        shape (None, None).
      actor_out: actor output sequences
      critic_out: critic output sequences. Should be of
        shape (None, None).
      states_out: resulting RNN states
    """
    def __init__(self, session, action_dist, obs_vectorizer):
        """
        Construct a recurrent model.
        """
        super(RecurrentAC, self).__init__(session, action_dist, obs_vectorizer)

        self.seq_lens_ph = tf.placeholder(tf.int32, shape=(None,))
        self.is_init_state_ph = tf.placeholder(tf.bool, shape=(None,))

        # Set this to a variable or a tuple of variables
        # for your model's initial state.
        self.init_state_vars = None

        # Set this to a placeholder for a batch of
        # observation sequences.
        self.obs_ph = None

        # Set this to a placeholder or a tuple of
        # placeholders for the first state in each
        # sequence.
        #
        # If a value in _is_init_state is True, then the
        # corresponding entry here is ignored.
        self.first_state_ph = None

        # Set this to a placeholder for a batch of mask
        # sequences to ignore timesteps in variable-length
        # sequences.
        self.mask_ph = None

        # Set these to the model outputs.
        self.actor_out = None
        self.critic_out = None
        self.states_out = None

    def scale_outputs(self, scale):
        """
        Scale the network outputs by the given amount.

        This may be called right after initializing the
        model to help deal with different reward scales.
        """
        self.critic_out *= scale
        self.actor_out *= scale

    @property
    def stateful(self):
        return True

    def start_state(self, batch_size):
        if isinstance(self.init_state_vars, tuple):
            res = []
            # pylint: disable=E1133
            for var in self.init_state_vars:
                var_val = self.session.run(var)
                res.append(np.array([var_val] * batch_size))
            return tuple(res)
        var_val = self.session.run(self.init_state_vars)
        return np.array([var_val] * batch_size)

    def step(self, observations, states):
        vec_obs = self.obs_vectorizer.to_vecs(observations)
        feed_dict = {
            self.seq_lens_ph: [1] * len(observations),
            self.is_init_state_ph: [False] * len(observations),
            self.obs_ph: [[x] for x in vec_obs],
            self.mask_ph: [[1]] * len(observations)
        }

        if isinstance(self.first_state_ph, tuple):
            assert isinstance(states, tuple)
            for key, value in zip(self.first_state_ph, states):
                feed_dict[key] = value
        else:
            feed_dict[self.first_state_ph] = states

        acts, vals, states = self.session.run((self.actor_out,
                                               self.critic_out,
                                               self.states_out),
                                              feed_dict)
        action_params = [a[0] for a in acts]
        return {
            'action_params': action_params,
            'actions': self.action_dist.sample(action_params),
            'states': states,
            'values': np.array(vals).flatten()
        }

    def batch_outputs(self):
        seq_shape = tf.shape(self.actor_out)
        out_count = seq_shape[0] * seq_shape[1]
        actor_shape = (out_count,) + self.action_dist.param_shape
        critic_shape = (out_count,)
        return (tf.reshape(self.actor_out, actor_shape),
                tf.reshape(self.critic_out, critic_shape),
                tf.reshape(self.mask_ph, critic_shape))

    # pylint: disable=R0914
    def batches(self, rollouts, batch_size=None):
        sizes = [r.num_steps for r in rollouts]
        for rollout_indices in mini_batches(sizes, batch_size):
            batch = [rollouts[i] for i in rollout_indices]
            max_len = max([r.num_steps for r in batch])
            obs_seqs = []
            is_inits = []
            masks = []
            rollout_idxs = []
            timestep_idxs = []
            for rollout_idx, rollout in zip(rollout_indices, batch):
                obs_seq = list(self.obs_vectorizer.to_vecs(rollout.step_observations))
                empty_obs = np.zeros(np.array(obs_seq[0]).shape)
                obs_seqs.append(_pad(obs_seq, max_len, value=empty_obs))
                is_inits.append(not rollout.trunc_start)
                masks.append(_pad([1]*rollout.num_steps, max_len))
                rollout_idxs.extend(_pad([rollout_idx]*rollout.num_steps, max_len))
                timestep_idxs.extend(_pad(list(range(rollout.num_steps)), max_len))
            feed_dict = {
                self.obs_ph: obs_seqs,
                self.is_init_state_ph: is_inits,
                self.mask_ph: masks,
                self.seq_lens_ph: [r.num_steps for r in batch]
            }
            self._add_first_states(feed_dict, batch)
            yield {
                'rollout_idxs': rollout_idxs,
                'timestep_idxs': timestep_idxs,
                'feed_dict': feed_dict
            }

    def create_state_fields(self, dtype, state_size):
        """
        Set self.first_state_ph and self.init_state_vars
        with the given TF datatype and the state size.

        The state size may be an integer, a TensorShape,
        or a tuple thereof.
        """
        if isinstance(state_size, tuple):
            self.first_state_ph = ()
            self.init_state_vars = ()
            for sub_shape in state_size:
                placeholder = tf.placeholder(dtype, _add_outer_none(sub_shape))
                variable = tf.Variable(tf.zeros(sub_shape))
                self.first_state_ph += (placeholder,)
                self.init_state_vars += (variable,)
        else:
            placeholder = tf.placeholder(dtype, _add_outer_none(state_size))
            variable = tf.Variable(tf.zeros(state_size))
            self.first_state_ph = placeholder
            self.init_state_vars = variable

    def _add_first_states(self, feed_dict, rollouts):
        """
        Add first state placeholders for the rollouts.
        """
        if isinstance(self.init_state_vars, tuple):
            for i, placeholder in enumerate(self.first_state_ph):
                first_states = [r.start_state[i][0] for r in rollouts]
                feed_dict[placeholder] = first_states
        else:
            first_states = [r.start_state[0] for r in rollouts]
            feed_dict[self.first_state_ph] = first_states

class RNNCellAC(RecurrentAC):
    """
    A recurrent actor-critic that uses a TensorFlow
    RNNCell.

    For RNNCells with nested tuple states, the tuple is
    flattened.
    """
    # pylint: disable=R0913
    # pylint: disable=R0914
    def __init__(self, session, action_dist, obs_vectorizer, make_cell):
        super(RNNCellAC, self).__init__(session, action_dist, obs_vectorizer)
        obs_seq_shape = (None, None) + obs_vectorizer.out_shape
        self.obs_ph = tf.placeholder(tf.float32, obs_seq_shape)
        self.mask_ph = tf.placeholder(tf.float32, (None, None))

        with tf.variable_scope('cell_input'):
            cell_input = self.cell_input_sequences()

        with tf.variable_scope('cell'):
            cell = make_cell()
        with tf.variable_scope('states'):
            if isinstance(cell.state_size, tuple):
                self.create_state_fields(tf.float32, tuple(nest.flatten(cell.state_size)))
            else:
                self.create_state_fields(tf.float32, cell.state_size)
        init_state = mix_init_states(self.is_init_state_ph,
                                     self.init_state_vars,
                                     self.first_state_ph)
        with tf.variable_scope('base'):
            if isinstance(init_state, tuple):
                init_state = nest.pack_sequence_as(cell.state_size, init_state)
            self.base_out, states = tf.nn.dynamic_rnn(cell, cell_input,
                                                      sequence_length=self.seq_lens_ph,
                                                      initial_state=init_state)
        if isinstance(states, tuple):
            self.states_out = tuple(nest.flatten(states))
        else:
            self.states_out = states
        with tf.variable_scope('actor'):
            self.actor_out = self.actor(self.base_out)
        with tf.variable_scope('critic'):
            self.critic_out = self.critic(self.base_out)

    def cell_input_sequences(self):
        """
        Transform the observation sequence placeholder
        into a batch of sequences for the RNN cell.

        Default behavior is to flatten the observation
        sequence.
        """
        batch_size = tf.shape(self.obs_ph)[0]
        seq_len = tf.shape(self.obs_ph)[1]
        obs_size = product(self.obs_vectorizer.out_shape)
        flat_shape = (batch_size, seq_len, obs_size)
        return tf.reshape(self.obs_ph, flat_shape)

    def actor(self, cell_outputs):
        """
        Transform the cell outputs into action parameters.

        The default behavior is to use a zero
        fully-connected layer with no activation.
        """
        zeros = tf.zeros_initializer()
        out_size = product(self.action_dist.param_shape)
        actor_out_seq = fully_connected(cell_outputs, out_size,
                                        activation_fn=None,
                                        weights_initializer=zeros)
        seq_shape = tf.shape(actor_out_seq)
        new_shape = ((seq_shape[0], seq_shape[1]) +
                     self.action_dist.param_shape)
        return tf.reshape(actor_out_seq, new_shape)

    # pylint: disable=R0201
    def critic(self, cell_outputs):
        """
        Transform the cell outputs into value predictions.

        The default behavior is to use a fully-connected
        layer with no activation.
        """
        zeros = tf.zeros_initializer()
        raw = fully_connected(cell_outputs, 1,
                              activation_fn=None,
                              weights_initializer=zeros)
        shape = tf.shape(raw)
        return tf.reshape(raw, (shape[0], shape[1]))

def _pad(unpadded, length, value=0):
    """
    Pad the list with the given value.
    """
    return unpadded + [value] * (length - len(unpadded))

def _add_outer_none(shape):
    """
    Add None as an outer dimension for the shape.
    """
    if isinstance(shape, tf.TensorShape):
        return [None] + shape.dims
    return [None, shape]



"""
Proximal policy optimization.
"""

import tensorflow as tf

from .a2c import A2C
from . import util

# pylint: disable=R0902
class PPO(A2C):
    """
    Train TensorFlow actor-critic models using PPO.

    This works with any model that implements
    anyrl.models.TFActorCritic.

    For more on PPO, see:
    https://arxiv.org/abs/1707.06347
    """
    def __init__(self,
                 model,
                 epsilon=0.2,
                 **a2c_kwargs):
        self._epsilon = epsilon
        param_shape = (None,) + model.action_dist.param_shape
        self._orig_action_params = tf.placeholder(tf.float32, param_shape)
        super(PPO, self).__init__(model, **a2c_kwargs)

    def _create_objective(self, vf_coeff, entropy_reg):
        actor, critic, mask = self.model.batch_outputs()

        dist = self.model.action_dist
        new_log_probs = dist.log_prob(actor, self._actions)
        old_log_probs = dist.log_prob(self._orig_action_params, self._actions)
        clipped_obj = clipped_objective(new_log_probs, old_log_probs,
                                        self._advs, self._epsilon)
        clipped_samples = _clipped_samples(new_log_probs, old_log_probs,
                                           self._advs, self._epsilon)
        critic_error = self._target_vals - critic
        self.actor_loss = -util.masked_mean(mask, clipped_obj)
        self.critic_loss = util.masked_mean(mask, tf.square(critic_error))
        self.entropy = util.masked_mean(mask, dist.entropy(actor))
        self.num_clipped = tf.cast(util.masked_sum(mask, clipped_samples), tf.int32)
        self.objective = (entropy_reg * self.entropy - self.actor_loss -
                          vf_coeff * self.critic_loss)

    def feed_dict(self, rollouts, batch=None, advantages=None, targets=None):
        if batch is None:
            batch = next(self.model.batches(rollouts))
        feed_dict = super(PPO, self).feed_dict(rollouts,
                                               batch=batch,
                                               advantages=advantages,
                                               targets=targets)
        orig_outs = util.select_model_out_from_batch('action_params', rollouts, batch)
        feed_dict[self._orig_action_params] = orig_outs
        return feed_dict

    # pylint: disable=W0221
    def optimize(self, learning_rate=1e-3):
        trainer = tf.train.AdamOptimizer(learning_rate=learning_rate)
        return trainer.minimize(-self.objective)

    # pylint: disable=R0913
    def run_optimize(self, optimize_op, rollouts, batch_size=None, num_iter=12,
                     log_fn=None, extra_feed_dict=None):
        """
        Run several steps of training with mini-batches.

        If log_fn is set, intermediate progress is logged
        by calling log_fn with string log messages.
        """
        batch_idx = 0
        batches = self.model.batches(rollouts, batch_size=batch_size)
        advantages = self.adv_est.advantages(rollouts)
        targets = self.adv_est.targets(rollouts)
        for batch in batches:
            terms = (self.actor_loss, self.critic_loss, self.entropy,
                     self.num_clipped, optimize_op)
            feed_dict = self.feed_dict(rollouts, batch,
                                       advantages=advantages,
                                       targets=targets)
            if extra_feed_dict:
                feed_dict.update(extra_feed_dict)
            terms = self.model.session.run(terms, feed_dict)
            if log_fn is not None:
                log_fn('batch %d: actor=%f critic=%f entropy=%f clipped=%d' %
                       (batch_idx, -terms[0], terms[1], terms[2], terms[3]))
            batch_idx += 1
            if batch_idx == num_iter:
                break

    # TODO: API that supports schedules and runs the
    # entire training loop for us.

def clipped_objective(new_log_probs, old_log_probs, advs, epsilon):
    """
    Compute the component-wise clipped PPO objective.
    """
    prob_ratio = tf.exp(new_log_probs - old_log_probs)
    clipped_ratio = tf.clip_by_value(prob_ratio, 1-epsilon, 1+epsilon)
    return tf.minimum(advs*clipped_ratio, advs*prob_ratio)

def _clipped_samples(new_log_probs, old_log_probs, advs, epsilon):
    """
    Count the number of samples that are clipped by the
    clipped PPO objective.
    """
    prob_ratio = tf.exp(new_log_probs - old_log_probs)
    clipped_ratio = tf.clip_by_value(prob_ratio, 1-epsilon, 1+epsilon)
    return tf.cast(advs*clipped_ratio < advs*prob_ratio, tf.float32)



"""
APIs for continuous spaces.
"""

import math

import numpy as np
import tensorflow as tf

from .base import Distribution

class BoxGaussian(Distribution):
    """
    A probability distribution over continuous variables,
    parameterized as a diagonal gaussian.
    """
    def __init__(self, low, high):
        self.low = low
        self.high = high

    @property
    def out_shape(self):
        return self.low.shape

    def to_vecs(self, space_elements):
        return np.array(space_elements)

    @property
    def param_shape(self):
        return self.low.shape + (2,)

    def sample(self, param_batch):
        params = np.array(param_batch)
        means, log_stddevs = self._mean_and_log_stddevs(params)
        stddevs = np.exp(log_stddevs)
        return np.random.normal(loc=means, scale=stddevs)

    def mode(self, param_batch):
        params = np.array(param_batch)
        return self._mean_and_log_stddevs(params)[0]

    def log_prob(self, param_batch, sample_vecs):
        means, log_stddevs = self._mean_and_log_stddevs(param_batch)
        constant_factor = 0.5 * math.log(2 * math.pi)
        diff = 0.5 * tf.square((means - sample_vecs) / tf.exp(log_stddevs))
        neg_log_probs = constant_factor + log_stddevs + diff
        return _reduce_sums(tf.negative(neg_log_probs))

    def entropy(self, param_batch):
        _, log_stddevs = self._mean_and_log_stddevs(param_batch)
        constant_factor = 0.5 * (math.log(2 * math.pi) + 1)
        return _reduce_sums(constant_factor + log_stddevs)

    def kl_divergence(self, param_batch_1, param_batch_2):
        means_1, log_stddevs_1 = self._mean_and_log_stddevs(param_batch_1)
        means_2, log_stddevs_2 = self._mean_and_log_stddevs(param_batch_2)
        # log(s2/s1) + (s1^2 + (u1 - u2)^2)/(2*s2^2) - 0.5
        term_1 = log_stddevs_2 - log_stddevs_1
        term_2_num = tf.exp(2 * log_stddevs_1) + tf.square(means_1 - means_2)
        term_2_denom = 2 * tf.exp(2 * log_stddevs_2)
        return _reduce_sums(term_1 + term_2_num/term_2_denom - 0.5)

    def _mean_and_log_stddevs(self, param_batch):
        """
        Compute the means and variances for a batch of
        parameters.
        """
        means = param_batch[..., 0]
        log_stddevs = param_batch[..., 1]
        bias = (self.high + self.low) / 2
        scale = (self.high - self.low) / 2
        return means + bias, log_stddevs + np.log(scale)

class BoxBeta(Distribution):
    """
    A probability distribution over continuous variables,
    parameterized as a component-wise scaled beta
    distribution.

    By default, inputs to the distribution are fed through
    `1 + softplus` to ensure that they are valid.
    """
    def __init__(self, low, high, softplus=True):
        self.low = low
        self.high = high
        self.softplus = softplus

    @property
    def out_shape(self):
        return self.low.shape

    def to_vecs(self, space_elements):
        return np.array(space_elements)

    @property
    def param_shape(self):
        return self.low.shape + (2,)

    def sample(self, param_batch):
        params = self._squash_inputs(np.array(param_batch))
        raw = np.random.beta(params[..., 0], params[..., 1])
        return raw * (self.high - self.low) + self.low

    def mode(self, param_batch):
        params = self._squash_inputs(np.array(param_batch))
        alpha, beta = params[..., 0], params[..., 1]
        raw = (alpha - 1) / (alpha + beta - 2)
        return raw * (self.high - self.low) + self.low

    def log_prob(self, param_batch, sample_vecs):
        scaled_samples = (sample_vecs - self.low) / (self.high - self.low)
        epsilon = 1e-20
        scaled_samples = tf.clip_by_value(scaled_samples, 0+epsilon, 1-epsilon)
        raw_probs = self._create_dist(param_batch).log_prob(scaled_samples)
        return _reduce_sums(raw_probs - np.log(self.high - self.low))

    def entropy(self, param_batch):
        raw_ents = self._create_dist(param_batch).entropy()
        return _reduce_sums(raw_ents + np.log(self.high - self.low))

    def kl_divergence(self, param_batch_1, param_batch_2):
        # KL is scale-invariant.
        return _reduce_sums(tf.contrib.distributions.kl_divergence(
            self._create_dist(param_batch_1),
            self._create_dist(param_batch_2)))

    def _create_dist(self, param_batch):
        params = self._squash_inputs(param_batch)
        return tf.contrib.distributions.Beta(params[..., 0], params[..., 1])

    def _squash_inputs(self, inputs):
        if not self.softplus:
            return inputs
        if isinstance(inputs, np.ndarray):
            softplus = np.log(1 + np.exp(inputs))
            non_linear = (inputs < 30)
            return 1 + np.where(non_linear, softplus, inputs)
        return 1 + tf.nn.softplus(inputs)

def _reduce_sums(batch):
    """
    Reduce a batch of shape [batch x out_shape] to a
    batch of scalars.
    """
    dims = list(range(1, len(batch.shape)))
    return tf.reduce_sum(batch, axis=dims)



"""
APIs for logging episode rollouts.
"""

import os
import time

import pandas

class EpisodeLogger:
    """
    A handle to an episode log file.

    Logs are CSV files with three columns:
      r: episode reward
      l: episode length (timesteps)
      t: timestamp of episode end, relative to log start.
    """
    def __init__(self, log_path):
        self._start_time = time.time()
        self._columns = ['r', 'l', 't']
        if os.path.isfile(log_path):
            contents = pandas.read_csv(log_path)
            # pylint: disable=C1801
            if len(contents) > 0:
                self._start_time = time.time() - max(contents['t'])
            self._out_file = open(log_path, 'a')
        else:
            self._out_file = open(log_path, 'wt+')
            self._write_header()

    def write_rollouts(self, rollouts):
        """
        Log the completed episodes from the rollouts.
        """
        data = {'r': [], 'l': [], 't': []}
        for rollout in rollouts:
            if rollout.trunc_end:
                continue
            data['r'].append(rollout.total_reward)
            data['l'].append(rollout.total_steps)
            data['t'].append(rollout.end_time - self._start_time)
        # pylint: disable=C1801
        if not data['r']:
            return
        self.write_frame(pandas.DataFrame(data))

    def write_frame(self, frame):
        """
        Write the pandas DataFrame to the end of the log.

        The frame must have r, l, and t columns.
        """
        new_frame = frame.sort_values(by='t')
        data = new_frame.to_csv(columns=self._columns, header=False, index=False)
        self._out_file.write(data)
        self._out_file.flush()

    def close(self):
        """
        Close the log file.
        """
        self._out_file.close()

    def _write_header(self):
        """
        Write the CSV header.
        """
        self._out_file.write(','.join(self._columns) + '\n')
        self._out_file.flush()

    def __enter__(self):
        return self

    def __exit__(self, _, value, traceback):
        self.close()



"""
Environments that wrap and modify other environments.
"""

from .batched import BatchedWrapper, BatchedFrameStack
from .image import DownsampleEnv, FrameStackEnv, GrayscaleEnv, MaxEnv, ResizeImageEnv
from .logs import LoggedEnv
from .meta import RL2Env, SwitchableEnv
from .padding import ObservationPadEnv, MultiBinaryPadEnv



"""
Reward normalization schemes.
"""

from math import sqrt

import numpy as np

# pylint: disable=R0903
class RewardNormalizer:
    """
    Normalize rewards in rollouts with a gradually
    updating divisor.
    """
    def __init__(self, update_rate=0.05, epsilon=1e-5):
        """
        Create a reward normalizer.

        Arguments:
          update_rate: the speed at which the normalizing
            coefficient updates (0 through 1). Set to None
            to use a running average over all rewards.
          epsilon: used to avoid dividing by 0
        """
        self._average = OnlineAverage(rate=update_rate)
        self._epsilon = epsilon
        self._coeff = None

    def update(self, rollouts):
        """
        Update the statistics using the rollouts and
        return a normalized copy of the rollouts.
        """
        squares = [rew**2 for r in rollouts for rew in r.rewards]
        self._average.update(squares)
        return [self._normalized_rollout(r) for r in rollouts]

    def _normalized_rollout(self, rollout):
        """
        Normalize the rollout.
        """
        scale = 1 / (self._epsilon + sqrt(self._average.value))
        rollout = rollout.copy()
        rollout.rewards = [r*scale for r in rollout.rewards]
        return rollout

class OnlineAverage:
    """
    A moving or running average.

    Running averages are unbiased and compute the mean of
    a list of values, even if those values come in in a
    stream.

    Moving averages are biased towards newer values,
    updating in a way that forgets the distant past.
    """
    def __init__(self, rate=None):
        """
        Create a new OnlineAverage.

        Args:
          rate: the moving average update rate. Used in
            update as `rate*(new-old)`, where new is the
            average of a new batch. If None, a dynamic
            update rate is chosen such that the online
            average is a running average over all the
            samples.
        """
        self.rate = rate
        self._current = 0
        self._num_samples = 0

    @property
    def value(self):
        """
        Get the current average value.
        """
        return self._current

    def update(self, values):
        """
        Update the moving average with the value batch.

        Args:
          values: a sequence of numerics.

        Returns:
          The new online average.
        """
        if self.rate is not None:
            rate = self.rate
            if self._num_samples == 0:
                rate = 1
        else:
            rate = len(values) / (len(values) + self._num_samples)
        self._current += rate * (np.mean(values) - self._current)
        self._num_samples += len(values)
        return self._current



"""
APIs for binary and multi-binary spaces.
"""

import numpy as np
import tensorflow as tf

from .base import Distribution

class MultiBernoulli(Distribution):
    """
    A multi-dimensional binary distribution parameterized
    as Bernoulli distributions with probabilities from
    sigmoids.
    """
    def __init__(self, num_bits):
        self.num_bits = num_bits

    @property
    def out_shape(self):
        return (self.num_bits,)

    def to_vecs(self, space_elements):
        return np.array(space_elements)

    @property
    def param_shape(self):
        return (self.num_bits,)

    def sample(self, param_batch):
        probs = 1 / (1 + np.exp(np.negative(np.array(param_batch))))
        rand = np.random.uniform(size=probs.shape)
        return (probs > rand).astype('int')

    def mode(self, param_batch):
        probs = 1 / (1 + np.exp(np.negative(np.array(param_batch))))
        return (probs > 0.5).astype('int')

    def log_prob(self, param_batch, sample_vecs):
        sample_vecs = tf.cast(sample_vecs, param_batch.dtype)
        log_probs_on = tf.log_sigmoid(param_batch) * sample_vecs
        log_probs_off = tf.log_sigmoid(-param_batch) * (1-sample_vecs)
        return tf.reduce_sum(log_probs_on + log_probs_off, axis=-1)

    def entropy(self, param_batch):
        ent_on = tf.log_sigmoid(param_batch) * tf.sigmoid(param_batch)
        ent_off = tf.log_sigmoid(-param_batch) * tf.sigmoid(-param_batch)
        return tf.negative(tf.reduce_sum(ent_on + ent_off, axis=-1))

    def kl_divergence(self, param_batch_1, param_batch_2):
        probs_on = tf.sigmoid(param_batch_1)
        probs_off = tf.sigmoid(-param_batch_1)
        log_diff_on = tf.log_sigmoid(param_batch_1) - tf.log_sigmoid(param_batch_2)
        log_diff_off = tf.log_sigmoid(-param_batch_1) - tf.log_sigmoid(-param_batch_2)
        kls = probs_on*log_diff_on + probs_off*log_diff_off
        return tf.reduce_sum(kls, axis=-1)



"""
A set of APIs for parameterizing probability distributions
in Reinforcement Learning.
"""

from .base import Vectorizer, Distribution
from .categorical import CategoricalSoftmax, NaturalSoftmax
from .continuous import BoxGaussian, BoxBeta
from .binary import MultiBernoulli
from .aggregate import TupleDistribution
from .gym import gym_space_distribution, gym_space_vectorizer

__all__ = dir()



"""
Wrappers that act on batched environments.

These can be useful in cases where computations are more
efficient in batches in the parent process, rather than
individually in each environment sub-process.

For example, suppose you want to feed screen observations
through a pre-trained CNN before passing them to your RL
model. If you don't want to fine-tune this CNN, then it is
most efficient to make the CNN part of the environment.
It makes sense to do this CNN as a batched wrapper, and it
may even be desirable to use a batched framestack wrapper
on top of the CNN wrapper.
"""

import gym
import numpy as np

from ..base import BatchedEnv

class BatchedWrapper(BatchedEnv):
    """
    A BatchedEnv that, by default, forwards all calls to a
    wrapped BatchedEnv.
    """
    def __init__(self, env):
        self.env = env
        if hasattr(env, 'observation_space'):
            self.observation_space = env.observation_space
        if hasattr(env, 'action_space'):
            self.action_space = env.action_space

    @property
    def num_sub_batches(self):
        return self.env.num_sub_batches

    @property
    def num_envs_per_sub_batch(self):
        return self.env.num_envs_per_sub_batch

    def reset_start(self, sub_batch=0):
        self.env.reset_start(sub_batch=sub_batch)

    def reset_wait(self, sub_batch=0):
        return self.env.reset_wait(sub_batch=sub_batch)

    def step_start(self, actions, sub_batch=0):
        self.env.step_start(actions, sub_batch=sub_batch)

    def step_wait(self, sub_batch=0):
        return self.env.step_wait(sub_batch=sub_batch)

    def close(self):
        self.env.close()

class BatchedFrameStack(BatchedWrapper):
    """
    The batched analog of FrameStackEnv.
    """
    def __init__(self, env, num_images=2):
        super(BatchedFrameStack, self).__init__(env)
        if hasattr(self, 'observation_space'):
            old_space = self.observation_space
            self.observation_space = gym.spaces.Box(np.repeat(old_space.low, num_images, axis=-1),
                                                    np.repeat(old_space.high, num_images, axis=-1))
        self._num_images = num_images
        self._history = [None] * env.num_sub_batches

    def reset_wait(self, sub_batch=0):
        obses = super(BatchedFrameStack, self).reset_wait(sub_batch=sub_batch)
        self._history[sub_batch] = [[o]*self._num_images for o in obses]
        return self._packed_obs(sub_batch)

    def step_wait(self, sub_batch=0):
        obses, rews, dones, infos = super(BatchedFrameStack, self).step_wait(sub_batch=sub_batch)
        for i, (obs, done) in enumerate(zip(obses, dones)):
            if done:
                self._history[sub_batch][i] = [obs] * self._num_images
            else:
                self._history[sub_batch][i].append(obs)
                self._history[sub_batch][i] = self._history[sub_batch][i][1:]
        return self._packed_obs(sub_batch), rews, dones, infos

    def _packed_obs(self, sub_batch):
        """
        Pack the sub-batch's observation along the
        inner dimension.
        """
        return [np.concatenate(o, axis=-1) for o in self._history[sub_batch]]



"""
Utilities for distributed training with MPI.
"""

from mpi4py import MPI
import numpy as np
import tensorflow as tf

# pylint: disable=E1101

class MPIOptimizer:
    """
    Wraps a TensorFlow optimizer to use MPI allreduce.
    """
    def __init__(self, optimizer, loss, var_list=None):
        old_variables = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)
        self.grads = [pair for pair in optimizer.compute_gradients(loss, var_list=var_list)
                      if pair[0] is not None]

        # TODO: make sure gradients will be ordered
        # deterministically.

        self.placeholders = []
        apply_in = []
        for grad, var in self.grads:
            placeholder = tf.placeholder(dtype=grad.dtype, shape=grad.shape)
            self.placeholders.append(placeholder)
            apply_in.append((placeholder, var))
        self.apply = optimizer.apply_gradients(apply_in)
        self._optimizer_vars = [v for v in tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)
                                if v not in old_variables]

    def minimize(self, sess, feed_dict=None, terms=None):
        """
        Compute the gradients, aggregate them, and apply
        them using the wrapped optimizer.

        Arguments:
          sess: the TensorFlow session.
          feed_dict: the TensorFlow feed_dict for the
            objective.
          terms: a list of scalar Tensors to run at the
             same time as the gradient computation.

        Returns:
          A tuple containing the mean values for each
          entry in terms.
        """
        if not feed_dict:
            feed_dict = {}
        if not terms:
            terms = []
        outs = sess.run(terms + [x[0] for x in self.grads],
                        feed_dict=feed_dict)
        grad_outs = outs[len(terms):]
        term_outs = outs[:len(terms)]

        extra_feed = feed_dict.copy()
        for grad_out, placeholder in zip(grad_outs, self.placeholders):
            mean_grad = np.zeros(grad_out.shape, dtype='float32')
            send_grad = np.array(grad_out, dtype='float32')
            MPI.COMM_WORLD.Allreduce(send_grad, mean_grad, op=MPI.SUM)
            mean_grad /= MPI.COMM_WORLD.Get_size()
            extra_feed[placeholder] = mean_grad
        sess.run(self.apply, feed_dict=extra_feed)

        result = []
        for term in term_outs:
            total = MPI.COMM_WORLD.allreduce(term, op=MPI.SUM)
            result.append(total / MPI.COMM_WORLD.Get_size())
        return tuple(result)

    def sync_from_root(self, sess):
        """
        Send the root node's parameters to every worker.

        This may add nodes to the graph, so it should not
        be called an unbounded number of times.

        This synchronizes both the model parameters and
        any extra variables created by the optimizer.

        Arguments:
          sess: the TensorFlow session.
        """
        rank = MPI.COMM_WORLD.Get_rank()
        for var in [v for _, v in self.grads] + self._optimizer_vars:
            if rank == 0:
                MPI.COMM_WORLD.bcast(sess.run(var))
            else:
                sess.run(tf.assign(var, MPI.COMM_WORLD.bcast(None)))

# pylint: disable=R0913
def mpi_ppo(ppo, optimizer, rollouts, batch_size=None, num_iter=12, log_fn=None,
            extra_feed_dict=None):
    """
    Run the PPO inner loop with an MPI optimizer.

    If log_fn is set, logging is done on rank 0.
    """
    batch_idx = 0
    batches = ppo.model.batches(rollouts, batch_size=batch_size)
    advantages = ppo.adv_est.advantages(rollouts)
    targets = ppo.adv_est.targets(rollouts)
    for batch in batches:
        feed_dict = ppo.feed_dict(rollouts, batch,
                                  advantages=advantages,
                                  targets=targets)
        if extra_feed_dict:
            feed_dict.update(extra_feed_dict)
        terms = optimizer.minimize(ppo.model.session,
                                   feed_dict=feed_dict,
                                   terms=[ppo.actor_loss, ppo.critic_loss, ppo.entropy,
                                          ppo.num_clipped])
        if log_fn and MPI.COMM_WORLD.Get_rank() == 0:
            log_fn('batch %d: actor=%f critic=%f entropy=%f clipped=%d' %
                   (batch_idx, -terms[0], terms[1], terms[2], terms[3]))
        batch_idx += 1
        if batch_idx == num_iter:
            break



"""
Helpers for implementing models.
"""

import numpy as np
import tensorflow as tf

def mini_batches(size_per_index, batch_size=None):
    """
    Generate mini-batches of size batch_size.

    The size_per_index list is the size of each batch
    element.
    Batches are generated such that the sum of the sizes
    of the batch elements is at least batch_size.
    """
    if batch_size is None or sum(size_per_index) <= batch_size:
        while True:
            yield list(range(len(size_per_index)))
    cur_indices = []
    cur_size = 0
    for idx in _infinite_random_shuffle(len(size_per_index)):
        cur_indices.append(idx)
        cur_size += size_per_index[idx]
        if cur_size >= batch_size:
            yield cur_indices
            cur_indices = []
            cur_size = 0

def _infinite_random_shuffle(num_elements):
    """
    Continually permute the elements and yield all of the
    permuted indices.
    """
    while True:
        for elem in np.random.permutation(num_elements):
            yield elem

def mix_init_states(is_init, init_states, start_states):
    """
    Mix initial variables with start state placeholders.
    """
    if isinstance(init_states, tuple):
        assert isinstance(start_states, tuple)
        res = []
        for sub_init, sub_start in zip(init_states, start_states):
            res.append(mix_init_states(is_init, sub_init, sub_start))
        return tuple(res)
    batch_size = tf.shape(start_states)[0]
    return tf.where(is_init, _batchify(batch_size, init_states), start_states)

def _batchify(batch_size, tensor):
    """
    Repeat a tensor the given number of times in the outer
    dimension.
    """
    batchable = tf.reshape(tensor, tf.concat([[1], tf.shape(tensor)], axis=0))
    ones = tf.ones(tensor.shape.ndims, dtype=tf.int32)
    repeat_count = tf.concat([[batch_size], ones], axis=0)
    return tf.tile(batchable, repeat_count)

def product(vals):
    """
    Compute the product of values in a list-like object.
    """
    prod = 1
    for val in vals:
        prod *= val
    return prod



"""
Manipulate lists of rollouts.
"""

def mean_total_reward(rollouts):
    """
    Get the mean of the total rewards.
    """
    return sum([r.total_reward for r in rollouts]) / len(rollouts)

def mean_finished_reward(rollouts):
    """
    Get the mean of the total rewards in all completed
    episodes.
    """
    complete = [r for r in rollouts if not r.trunc_end]
    if not complete:
        return 0.0
    return sum([r.total_reward for r in complete]) / len(complete)



"""
Test for learning algorithm utilities.
"""

# pylint: disable=E1129

import unittest

from anyrl.algos.util import masked_mean
import numpy as np
import tensorflow as tf

class TestMask(unittest.TestCase):
    """
    Test the masked_mean() API.
    """
    def test_basic(self):
        """
        Test on a normal set of inputs.
        """
        mean = _masked_mean([0, 1, 0, 1, 1], [2.5, 7, 3, 0.8, 0.9])
        self.assertTrue(np.allclose(mean, np.mean([7, 0.8, 0.9])))

    def test_unusual_values(self):
        """
        Test masking unusual values like nan.
        """
        mean = _masked_mean([0, 1, 0, 1, 1, 0],
                            [np.inf, 7, np.nan, 0.8, 0.9, np.inf])
        self.assertTrue(np.allclose(mean, np.mean([7, 0.8, 0.9])))

def _masked_mean(mask, values):
    """
    Compute the masked mean using TensorFlow.
    """
    with tf.Graph().as_default():
        with tf.Session() as sess:
            mask_ph = tf.placeholder(tf.float32, shape=(None,))
            values_ph = tf.placeholder(tf.float32, shape=(None,))
            mean = masked_mean(mask_ph, values_ph)
            return sess.run(mean, feed_dict={mask_ph: np.array(mask),
                                             values_ph: np.array(values)})

if __name__ == '__main__':
    unittest.main()



"""
Wrappers for meta-learning.
"""

import gym
import gym.spaces as spaces
import numpy as np

class RL2Env(gym.Wrapper):
    """
    A wrapper which augments the observation space to
    include the action, reward, and done value.

    Creates a tuple observation space:
      (observation, action, reward, done).

    See: https://arxiv.org/abs/1611.02779.
    """
    def __init__(self, env, first_action, num_eps=1, warmup_eps=0):
        """
        Parameters:
          env: the environment to wrap.
          first_action: the action to include in the first
            observation.
          num_eps: episodes per meta-episode.
          warmup_eps: the number of episodes at the start
            of a meta-episode for which rewards are 0.
            Negative values are added to num_eps.
        """
        if warmup_eps < 0:
            warmup_eps += num_eps
        super(RL2Env, self).__init__(env)
        self.first_action = first_action
        self.observation_space = spaces.Tuple([
            env.observation_space,
            env.action_space,
            spaces.Box(low=-np.inf, high=np.inf, shape=(1,)),
            spaces.MultiBinary(1)
        ])
        self.num_eps = num_eps
        self.warmup_eps = warmup_eps
        self._done_eps = 0

    # pylint: disable=W0221
    def _reset(self, **kwargs):
        self._done_eps = 0
        obs = self.env.reset(**kwargs)
        return (obs, self.first_action, np.array([0.0]), np.array([0]))

    def _step(self, action):
        obs, rew, done, info = self.env.step(action)
        aug_obs = (obs, action, np.array([rew]), np.array([int(done)]))
        if self._done_eps < self.warmup_eps:
            rew = 0.0
        if done:
            self._done_eps += 1
            if self._done_eps < self.num_eps:
                done = False
                aug_obs = (self.env.reset(),) + aug_obs[1:]
        return aug_obs, rew, done, info

class SwitchableEnv(gym.Env):
    """
    An environment that proxies calls to another
    environment that can be swapped out at any time.

    This is useful in conjunction with RL2Env to swap out
    the environment after every meta-episode.
    """
    def __init__(self, first_env):
        self.env = first_env
        self.action_space = first_env.action_space
        self.observation_space = first_env.observation_space

    def switch_env(self, new_env):
        """
        Switch to a new environment.

        The new environment must have the same spaces as
        the old one.
        """
        self.env = new_env

    # pylint: disable=W0221
    def _reset(self, **kwargs):
        return self.env.reset(**kwargs)

    def _step(self, action):
        return self.env.step(action)

    def _render(self, mode='human', close=False):
        return self.env.render(mode=mode, close=close)

    def _seed(self, seed=None):
        return self.env.seed(seed=seed)



"""
Types and functions for manipulating (partial) rollouts in
RL environments.
"""

def empty_rollout(start_state, prev_steps=0, prev_reward=0):
    """
    Create a Rollout with no timesteps.
    """
    return Rollout(observations=[], model_outs=[], rewards=[],
                   start_state=start_state, prev_steps=prev_steps,
                   prev_reward=prev_reward)

# pylint: disable=R0902
class Rollout:
    """
    A sequence of observations, actions, and rewards that
    were recorded by running an agent in an environment.

    Model outputs are structured like the results of
    Model.step(), but they have a batch size of one.

    Rollouts may represent total episodes, but they may
    also represent a sliced version of the episode.
    The trunc_start and trunc_end properties indicate if
    the episode was truncated at the start or the end.
    When trunc_start is True, prev_reward and prev_steps
    indicate the amount of reward and number of steps in
    the episode before this Rollout.
    When trunc_end is true, there is an extra observation
    and model_outs entry.

    Rollouts may also have an end_time, which is a UNIX
    timestamp of when the rollout was done being made.
    """
    # pylint: disable=R0913
    def __init__(self, observations, model_outs, rewards, start_state,
                 prev_steps=0, prev_reward=0, infos=None, end_time=0):
        assert len(observations) == len(model_outs)
        assert len(rewards) <= len(observations)
        assert len(observations) <= len(rewards)+1
        if infos is None:
            infos = [{}] * len(rewards)
        assert len(infos) == len(rewards)
        self.observations = observations
        self.model_outs = model_outs
        self.rewards = rewards
        self.start_state = start_state
        self.prev_steps = prev_steps
        self.prev_reward = prev_reward
        self.infos = infos
        self.end_time = end_time

    @property
    def trunc_end(self):
        """
        Get whether or not the episode completed in this
        Rollout.
        """
        return len(self.observations) > self.num_steps

    @property
    def trunc_start(self):
        """
        Get whether or not steps were taken in the episode
        before this Rollout.
        """
        return self.prev_steps > 0

    @property
    def num_steps(self):
        """
        Get the total number of timesteps (not including
        the extra observation or previous timesteps for
        truncated episodes).
        """
        return len(self.rewards)

    @property
    def step_observations(self):
        """
        Returns a list of observations that does not
        include the trailing observation for truncated
        episodes.
        """
        return self.observations[:self.num_steps]

    @property
    def step_model_outs(self):
        """
        Returns a list of model outputs that does not
        include the trailing output for truncated
        episodes.
        """
        return self.model_outs[:self.num_steps]

    @property
    def total_reward(self):
        """
        Get the total reward so far in the episode.
        This includes rewards from previous segments of
        this episode.
        """
        return sum(self.rewards) + self.prev_reward

    @property
    def total_steps(self):
        """
        Get the total number of completed timesteps in the
        episode.
        This includes timesteps from previous segments of
        this episode, but it does not include the final
        observation for truncated episodes.
        """
        return self.num_steps + self.prev_steps

    def predicted_value(self, timestep):
        """
        Get the predicted value at the given timestep.

        This only works if the model generated values.
        """
        return self.model_outs[timestep]['values'][0]

    def copy(self):
        """
        Create a shallow copy of the rollout.
        """
        return Rollout(self.observations,
                       self.model_outs,
                       self.rewards,
                       self.start_state,
                       prev_steps=self.prev_steps,
                       prev_reward=self.prev_reward,
                       infos=self.infos,
                       end_time=self.end_time)



"""
Tests for probability distributions.
"""

# pylint: disable=E1129

import random
import unittest

import gym
import numpy as np
import tensorflow as tf

from anyrl.spaces import (CategoricalSoftmax, NaturalSoftmax, BoxGaussian, BoxBeta, MultiBernoulli,
                          TupleDistribution, gym_space_distribution)

# Number of times to run sample-based tests.
NUM_SAMPLE_TRIES = 3

class DistributionTester:
    """
    Various generic tests for Distributions.
    """
    def __init__(self, test, dist, batch_size=200000, prec=3e-2):
        self.prec = prec
        self.test = test
        self.dist = dist
        self.batch_size = batch_size
        self.session = None

    def test_all(self):
        """
        Run all generic tests.
        """
        np.random.seed(1337)
        with tf.Graph().as_default():
            self.session = tf.Session()
            with self.session:
                self.test_shapes()
                self.test_entropy()
                self.test_kl()
                self.test_mode()

    def test_shapes(self):
        """
        Test the reported shapes.
        """
        for elem in self.dist.out_shape:
            self.test.assertTrue(isinstance(elem, int))
        for elem in self.dist.param_shape:
            self.test.assertTrue(isinstance(elem, int))

        samples = self.dist.sample(self._random_params())
        self.test.assertEqual(len(samples), self.batch_size)
        sample_shape = np.array(self.dist.to_vecs(samples)).shape
        self.test.assertEqual(sample_shape, self._out_batch_shape())

        param_placeholder = tf.placeholder(tf.float32, self._param_batch_shape())
        sample_placeholder = tf.placeholder(tf.float32, self._out_batch_shape())
        feed_dict = {
            param_placeholder: self._random_params(),
            sample_placeholder: self.dist.to_vecs(samples)
        }
        log_probs = self.dist.log_prob(param_placeholder, sample_placeholder)
        entropies = self.dist.entropy(param_placeholder)
        kls = self.dist.kl_divergence(param_placeholder, param_placeholder)
        outs = self.session.run([log_probs, entropies, kls], feed_dict=feed_dict)
        scalar_out_shape = (self.batch_size,)
        for out in outs:
            self.test.assertEqual(np.array(out).shape, scalar_out_shape)

    def test_entropy(self):
        """
        Test entropy with a sample estimate.
        """
        param_placeholder = tf.placeholder(tf.float32, self._param_batch_shape())
        sample_placeholder = tf.placeholder(tf.float32, self._out_batch_shape())
        log_probs = self.dist.log_prob(param_placeholder, sample_placeholder)
        entropy = tf.reduce_mean(self.dist.entropy(param_placeholder))
        for i in range(NUM_SAMPLE_TRIES):
            params = self._random_params(use_numpy=(i%2 == 0))
            samples = self.dist.sample(params)
            feed_dict = {
                param_placeholder: params,
                sample_placeholder: self.dist.to_vecs(samples)
            }
            log_probs_out = np.array(self.session.run(log_probs,
                                                      feed_dict=feed_dict))
            entropy_out = self.session.run(entropy, feed_dict=feed_dict)

            est_entropy = -np.mean(log_probs_out)
            self.test.assertTrue(abs(1 - entropy_out/est_entropy) < self.prec)

    def test_kl(self):
        """
        Test KL divergence with a sample estimate.
        """
        param_1_placeholder = tf.placeholder(tf.float32, self._param_batch_shape())
        param_2_placeholder = tf.placeholder(tf.float32, self._param_batch_shape())
        sample_placeholder = tf.placeholder(tf.float32, self._out_batch_shape())
        log_probs_1 = self.dist.log_prob(param_1_placeholder, sample_placeholder)
        log_probs_2 = self.dist.log_prob(param_2_placeholder, sample_placeholder)
        real_kl = tf.reduce_mean(self.dist.kl_divergence(param_1_placeholder,
                                                         param_2_placeholder))
        for i in range(NUM_SAMPLE_TRIES):
            params = self._random_params(use_numpy=(i%2 == 0))
            samples = self.dist.sample(params)
            feed_dict = {
                param_1_placeholder: params,
                param_2_placeholder: self._random_params(),
                sample_placeholder: self.dist.to_vecs(samples)
            }
            log_probs_out_1 = np.array(self.session.run(log_probs_1,
                                                        feed_dict=feed_dict))
            log_probs_out_2 = np.array(self.session.run(log_probs_2,
                                                        feed_dict=feed_dict))
            kl_out = self.session.run(real_kl, feed_dict=feed_dict)
            self.test.assertTrue(kl_out >= 0)

            est_kl = np.mean(log_probs_out_1 - log_probs_out_2)
            self.test.assertTrue(abs(1 - kl_out/est_kl) < self.prec)

    def test_mode(self):
        """
        Test that the mode is truly the most likely.
        """
        param_batch = self._random_params(use_numpy=True)
        samples = tf.constant(self.dist.to_vecs(self.dist.sample(param_batch)))
        modes = self.dist.to_vecs(self.dist.mode(param_batch))
        modes = tf.constant(modes)
        param_batch = tf.constant(param_batch)

        sample_probs = self.session.run(self.dist.log_prob(param_batch, samples))
        mode_probs = self.session.run(self.dist.log_prob(param_batch, modes))
        self.test.assertTrue((sample_probs <= mode_probs).all())

    def _random_params(self, use_numpy=False):
        params = np.random.normal(size=self.dist.param_shape)
        batch = [params] * self.batch_size
        if use_numpy:
            batch = np.array(batch)
        return batch

    def _param_batch_shape(self):
        return (self.batch_size,) + self.dist.param_shape

    def _out_batch_shape(self):
        return (self.batch_size,) + self.dist.out_shape

class TestCategoricalSoftmax(unittest.TestCase):
    """
    Tests for the CategoricalSoftmax distribution.
    """
    def test_log_prob(self, dist_maker=CategoricalSoftmax):
        """
        Test log probabilities on a known case.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                params = tf.constant(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
                                     dtype=tf.float32)
                samples = tf.constant(np.array([[0, 0, 1],
                                                [1, 0, 0],
                                                [0, 1, 0]]),
                                      dtype=tf.float32)
                dist = dist_maker(3)
                log_probs = np.array(sess.run(dist.log_prob(params, samples)))
                expected = np.array([0.665241, 0.090031, 0.244728])
                diff = np.amax(expected - np.exp(log_probs))
                self.assertTrue(diff < 1e-4)

    def test_generic(self):
        """
        Run generic tests with DistributionTester.
        """
        dist = CategoricalSoftmax(7, low=2)
        tester = DistributionTester(self, dist)
        tester.test_all()

class TestNaturalSoftmax(unittest.TestCase):
    """
    Tests for the NaturalSoftmax distribution.
    """
    def test_log_prob(self):
        """
        Test log probabilities against CategoricalSoftmax.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                dist = NaturalSoftmax(7)
                params = tf.constant(np.random.normal(size=(15, 7)), dtype=tf.float64)
                sampled = tf.one_hot([random.randrange(7) for _ in range(15)], 7,
                                     dtype=tf.float64)
                actual = sess.run(dist.log_prob(params, sampled))
                expected = sess.run(CategoricalSoftmax(7).log_prob(params, sampled))
                self.assertTrue(np.allclose(actual, expected))

    def test_gradient_determinism(self):
        """
        Make sure that the gradient doesn't change from
        run to run.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                dist = NaturalSoftmax(7)
                params = tf.constant(np.random.normal(size=(15, 7)), dtype=tf.float64)
                sampled = tf.one_hot([random.randrange(7) for _ in range(15)], 7,
                                     dtype=tf.float64)
                batched_grad = tf.gradients(dist.log_prob(params, sampled), params)[0]
                first = sess.run(batched_grad)
                for _ in range(10):
                    self.assertTrue(np.allclose(first, sess.run(batched_grad)))

    def test_natural_gradient(self):
        """
        Test random natural gradient cases.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                for size in range(3, 9):
                    dist = NaturalSoftmax(size, epsilon=0)
                    softmax = CategoricalSoftmax(size)
                    param_row = tf.constant(np.random.normal(size=(size,)), dtype=tf.float64)
                    params = tf.stack([param_row])
                    one_hot = np.zeros((1, size))
                    one_hot[0, 1] = 1
                    samples = tf.constant(one_hot, dtype=tf.float64)
                    kl_div = softmax.kl_divergence(tf.stop_gradient(params), params)
                    hessian = sess.run(tf.hessians(kl_div, param_row)[0])
                    gradient = sess.run(tf.gradients(softmax.log_prob(params, samples),
                                                     params)[0][0])
                    expected = np.matmul(np.array([gradient]), np.linalg.pinv(hessian))[0]
                    actual = sess.run(tf.gradients(dist.log_prob(params, samples), params)[0][0])
                    self.assertTrue(np.allclose(actual, expected))

    def test_batched_gradient(self):
        """
        Test that batched gradients give the same results
        as single gradients.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                dist = NaturalSoftmax(7)
                params = tf.constant(np.random.normal(size=(15, 7)), dtype=tf.float64)
                sampled = tf.one_hot([random.randrange(7) for _ in range(15)], 7,
                                     dtype=tf.float64)
                batched_grad = tf.gradients(dist.log_prob(params, sampled), params)[0]
                single_grads = []
                for i in range(15):
                    sub_params = params[i:i+1]
                    prob = dist.log_prob(sub_params, sampled[i:i+1])
                    single_grads.append(tf.gradients(prob, sub_params)[0])
                single_grad = tf.concat(single_grads, axis=0)
                batched, single = sess.run((batched_grad, single_grad))
                self.assertEqual(batched.shape, single.shape)
                self.assertTrue(np.allclose(batched, single))

class TestBoxGaussian(unittest.TestCase):
    """
    Tests for the BoxGaussian distribution.
    """
    def test_generic(self):
        """
        Run generic tests with DistributionTester.
        """
        dist = BoxGaussian(np.array([[-3, 7, 1], [1, 2, 3]]),
                           np.array([[5, 7.1, 3], [2, 3.1, 4]]))
        tester = DistributionTester(self, dist)
        tester.test_all()

class TestBoxBeta(unittest.TestCase):
    """
    Tests for the BoxBeta distribution.
    """
    def test_generic_simple(self):
        """
        Run generic tests with an unscaled distribution.
        """
        dist = BoxBeta(np.array([0]), np.array([1]))
        tester = DistributionTester(self, dist)
        tester.test_all()

    def test_generic(self):
        """
        Run generic tests with DistributionTester.
        """
        dist = BoxBeta(np.array([[-3, 7, 1], [1, 2, 3]]),
                       np.array([[5, 7.1, 3], [2, 3.1, 4]]))
        tester = DistributionTester(self, dist, batch_size=400000)
        tester.test_all()

    def test_log_prob_simple(self):
        """
        Test log probs for a very simple distribution.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                dist = BoxBeta(np.array([0]), np.array([1]), softplus=False)
                actual = sess.run(dist.log_prob(np.array([[0.5, 0.5]]), np.array([0.5])))
                expected = np.array([_beta_log_prob(0.5, 0.5, 0.5)])
                self.assertTrue(np.allclose(actual, expected, atol=1e-4))

    def test_log_prob(self):
        """
        Test log probs for known situations.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                dist = BoxBeta(np.array([[0], [-2]]), np.array([[1], [3]]), softplus=False)
                actual = sess.run(dist.log_prob(np.array([[[0.1, 0.5]], [[0.3, 0.7]]]),
                                                np.array([[0.4], [-0.5]])))
                expected = np.array([_beta_log_prob(0.1, 0.5, 0.4),
                                     _beta_log_prob(0.3, 0.7, 0.3) - np.log(5)])
                self.assertTrue(np.allclose(actual, expected, atol=1e-4))

def _beta_log_prob(alpha, beta, value):
    with tf.Graph().as_default():
        with tf.Session() as sess:
            gamma_a, gamma_b, gamma_ab = sess.run((tf.lgamma(alpha), tf.lgamma(beta),
                                                   tf.lgamma(alpha + beta)))
    log_denom = gamma_a + gamma_b - gamma_ab
    log_num = np.log((value ** (alpha - 1)) * ((1 - value) ** (beta - 1)))
    return log_num - log_denom

class TestMultiBernoulli(unittest.TestCase):
    """
    Tests for the MultiBernoulli distribution.
    """
    def test_generic(self):
        """
        Run generic tests with DistributionTester.
        """
        dist = MultiBernoulli(3)
        tester = DistributionTester(self, dist)
        tester.test_all()

class TestTuple(unittest.TestCase):
    """
    Tests for tuples of distributions.
    """
    def test_unpack_shape(self):
        """
        Make sure that unpack_outs gives the correct
        shape.
        """
        box_dist = BoxGaussian(np.array([[-3, 7, 1], [1, 2, 3.7]]),
                               np.array([[5, 7.1, 1.5], [2, 2.5, 4]]))
        dist = TupleDistribution((MultiBernoulli(3), box_dist))
        vec = np.array([[0, 1, 0, 1, 2, 3, 4, 5, 6],
                        [1, 0, 1, 6, 5, 4, 3, 2, 1]])
        unpacked = dist.unpack_outs(vec)
        self.assertEqual(len(unpacked), 2)
        self.assertEqual(np.array(unpacked[0]).shape, (2, 3))
        self.assertEqual(np.array(unpacked[1]).shape, (2, 2, 3))

    def test_generic(self):
        """
        Run generic tests with DistributionTester.
        """
        box_dist = BoxGaussian(np.array([[-3, 7, 1], [1, 2, 3.7]]),
                               np.array([[5, 7.1, 1.5], [2, 2.5, 4]]))
        dist = TupleDistribution((MultiBernoulli(3), box_dist))
        tester = DistributionTester(self, dist)
        tester.test_all()

    def test_multi_discrete(self):
        """
        Basic tests for MultiDiscrete, which is internally
        implemented using a tuple.
        """
        space = gym.spaces.MultiDiscrete([[1, 3], [-1, 5]])
        dist = gym_space_distribution(space)
        tester = DistributionTester(self, dist)
        tester.test_all()
        for sample in dist.sample(np.zeros((50,) + dist.param_shape)):
            self.assertTrue(space.contains(sample))

if __name__ == '__main__':
    unittest.main()
